// dmp2midi v1.0 written by Gnter Nagler 1996 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#ifdef __MSDOS__
#define WRITE_BINARY  "wb"
#define READ_TEXT     "r"
#else
#define WRITE_BINARY  "w"
#define READ_TEXT     "r"
#endif

char* input = 0;
char* output = 0;

FILE* f = 0;  // input file
MidiWrite* write = 0;

int resolution = 96;
int program = 0;
float tempo = 120.0;

void usage()
{
  fprintf(stderr, "dmp2midi converts text generated by mididmp back to midi\n");
  fprintf(stderr, "usage: dmp2midi [-tempo #] [-program #] filename.dmp filename.mid\n");
  fprintf(stderr, "-tempo #:   general song tempo in beats/min (40-240)\n");
  fprintf(stderr, "-program #: instrument number (0-127)\n");
  exit(1);
}

char line[256];
long linenum = 0;

typedef struct
{
  long evtime;
  unsigned char cmd, note, vel;
} event;

#define MAXEVENTS  1000
int events = 0;
event eventqueue[MAXEVENTS];

void writeevents(long untiltime = 0x7fffffffl)
{
int removed;

  do
  {
    removed = 0;
    int besti = -1;

    for (int i = 0; i < events; i++)
    if (eventqueue[i].evtime >= 0)
    {
      if (eventqueue[i].evtime <= untiltime)
      {
	if (besti < 0 || eventqueue[besti].evtime > eventqueue[i].evtime)
	  besti = i;
      }
    }
    if (besti >= 0)
    {
    assert(eventqueue[besti].evtime >= write->getcurunit());
      write->time(eventqueue[besti].evtime - write->getcurunit());

      if ((eventqueue[besti].cmd & 0xF0) == 0x80)
	write->noteoff(eventqueue[besti].cmd & 15, eventqueue[besti].note, eventqueue[besti].vel);
      else if ((eventqueue[besti].cmd & 0xF0) == 0x90)
	write->noteon(eventqueue[besti].cmd & 15, eventqueue[besti].note, eventqueue[besti].vel);
      else
	assert(0); // command not supported

      eventqueue[besti].evtime = -1;
      if (besti == events-1)
	events--;
      removed = 1;
    }
  } while (removed);
  while (events > 0 && eventqueue[events-1].evtime < 0)
    events--;
}

void addnote(long evtime, unsigned char cmd, unsigned char note, unsigned char vel)
{
  for (int i = 0; i < events; i++)
  if (eventqueue[i].evtime < 0)
    break;
  if (i >= events)
  {
    assert(events < MAXEVENTS);
    if (events >= MAXEVENTS)
      return;
    i = events++;
  }
  eventqueue[i].evtime = evtime;
  eventqueue[i].cmd = cmd;
  eventqueue[i].note = note;
  eventqueue[i].vel = vel;
}

int noteval(char* s)
{
int level = 0;

  s[0] = toupper(s[0]);
  switch(s[0])
  {
  case 'A': level = +9; break;
  case 'B':
  case 'H': level = +11; break;
  case 'C': level = 0; break;
  case 'D': level = 2; break;
  case 'E': level = 4; break;
  case 'F': level = 5; break;
  case 'G': level = 7; break;
  default: return -1;
  }
  s++;
  if (*s =='#')
  {
    level++;
    s++;
  }
  else if (strncmp(s, "is", 2) == 0)
  {
    level++;
    s+=2;
  }
  else if (*s == 'b' || *s == 's')
  {
    level--;
    s++;
  }
  else if (strncmp(s, "es", 2) == 0)
  {
    level--;
    s+=2;
  }

  if (*s >= '0' && *s <= '9')
    level += 12 * (atoi(s)+1);
  else
    level += 12 * 5;  // default octave
  return level;
}

int nextline()
{
  if (!fgets(line, sizeof(line)-1, f))
  {
    line[0] = '\0';
    return 0;
  }
  char* eoln = strchr(line, '\n');
  if (eoln)
    *eoln = '\0';
  linenum++;
  fprintf(stderr, "%ld\r", linenum);
  return 1;
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-tempo", 2) == 0)
    {
      argc--; argv++;
      if (argc == 0 || sscanf(*argv, "%f", &tempo) != 1)
      {
	fprintf(stderr, "option -tempo needs a value\n");
	continue;
      }
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-program", 2) == 0)
    {
      argc--; argv++;
      if (argc == 0 || sscanf(*argv, "%d", &program) != 1)
      {
	fprintf(stderr, "option -program needs a value 0-127\n");
	continue;
      }
      argc--; argv++; continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert dmp to same file\n");
    return 1;
  }

  f = fopen(input, READ_TEXT);
  if (!f)
  {
    perror(input);
    return 1;
  }
  if (!nextline() || // empty file?
      strstr(line, "units per quarternote") == 0 || // first line not matching
      sscanf(line, "%d units per quarternote", &resolution) != 1) // resolution missing
  {
    fprintf(stderr, "%s is not a valid dmp generated by mididmp\n");
    fclose(f);
    return 1;
  }
  write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return 1;
  }
  if (!write->getf())
  {
    perror(output);
    return 1;
  }
  write->head(/* version */ 1, /* track count updated later */0, resolution);

  write->track(); // tempo track
  write->tempo((long)(60000000.0 / tempo));
  write->tact(4,4,24,8);
  write->endtrack(); // tempo track

  nextline();
  long curbeattime = 0;
  long curbeatlen = 0;

  int curtactnom = 4;
  int curtactdenom = 4;

  do
  {
    if (line[0] == '\0')
    {
      if (!nextline())
	break;
      continue;
    }
    if (strncmp(line, "Track ", 6) == 0)
    {
      writeevents(); // write all remaining events of last track
      write->track();
      write->volume(0, 100);
      write->program(0, program);
      nextline();
      curbeattime = 0;
      curbeatlen = 0;
      continue;
    }
    else if (line[0] == 'M')
    {
      if (strstr(line, "Beat"))
      {
	int measure;
	int tactnom, tactdenom;

	char* lp = strstr(line, "Beat: ");
	if (sscanf(line+1, "%d", &measure) == 1 && lp && sscanf(lp+6, "%d/%d", &tactnom, &tactdenom) == 2 && tactdenom > 0 && tactnom > 0)
	{
	  curbeattime += curbeatlen;
	  curbeatlen = resolution * 4 * tactnom / tactdenom;

	  if (curtactnom != tactnom || curtactdenom != tactdenom)
	  {
	    write->tact(tactnom, tactdenom, 24, 8);
	    curtactnom = tactnom;
	    curtactdenom = tactdenom;
	  }
	  writeevents(curbeattime);
	  nextline();
	  continue;
	}
      }
      static int measure, atquarternote, atunit, velocity, quarternotes, units;
      static char notename[20];
      if (sscanf(line, "M%d	%d.%d  %s V%d   %d.%d", &measure, &atquarternote, &atunit, notename,  &velocity, &quarternotes, &units) == 7)
      {
	  long eventtime = curbeattime + (atquarternote-1) * resolution + atunit;
	  long eventlen = quarternotes * resolution + units;
	  int note = noteval(notename);

	  addnote(eventtime, 0x90, note, velocity); // event on
	  addnote(eventtime + eventlen, 0x80, note, velocity); // event off
	  nextline();
	  continue;
      }
      else if (sscanf(line, "M%d	%d.%d  %s V%d   TIE", &measure, &atquarternote, &atunit, notename,  &velocity) == 5)
      {
	  long eventtime = curbeattime + (atquarternote-1) * resolution + atunit;
	  int note = noteval(notename);

	  addnote(eventtime, 0x90, note, velocity); // event on
	  nextline();
	  continue;
      }
      else if (sscanf(line, "M%d	TIE  %s V%d   %d.%d", &measure, notename,  &velocity, &quarternotes, &units) == 5)
      {
	  int note = noteval(notename);
	  addnote(curbeattime + quarternotes*resolution + units, 0x80, note, velocity);
	  nextline();
	  continue;
      }
      else if (sscanf(line, "M%d	TIE  %s V%d   TIE", &measure, notename,  &velocity) == 3)
      {
	  nextline();
	  continue;
      }
    }
    fprintf(stderr, "Line %ld: invalid line ignored (%s)\n", linenum, line);
    if (!nextline())
      break;
  } while (1);
  writeevents(); // write all remaining events
  delete write;
  fclose(f);
  return 0;
}
