// midi0to1 v1.0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

char* input = 0;
char* output = 0;

int srcformat = 0;
int channelused[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

MidiWrite* write = 0;


class MidiChannelTrack : public MidiRead
{
public:
  MidiChannelTrack(char* name, int channel);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void time(unsigned long ticks);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void error(const char* msg);

  int channel_;
};

MidiChannelTrack::MidiChannelTrack(char* name, int channel) : MidiRead(name)
{
  options_ = OPTION_NOCONTROLS+OPTION_NOSYSEVENTS;
  channel_ = channel;
}

void MidiChannelTrack::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  if (channel_ == NOCHANNEL)
  {
    if (version == 1)
    {
      fprintf(stderr, "%s is already format 1\n", midiname_);
      exit(0);
    }
    if (version != 0)
    {
      fprintf(stderr, "%s is format %d. This program does not convert it to 1.\n", midiname_, version);
      exit(1);
    }
    srcformat = version;
    write->head(1, 0, unitsperbeat);
  }
}

void MidiChannelTrack::track(int trackno, long length, int channel)
{
  if (trackno == 1)
    write->track();
}

void MidiChannelTrack::endtrack(int trackno)
{
  if (trackno == 1)
    write->endtrack();
}

void MidiChannelTrack::meta(int what, int len, unsigned char* data)
{
  if (channel_ == NOCHANNEL)
    write->meta(what, len, data);
}

void MidiChannelTrack::time(unsigned long ticks)
{
  write->time(ticks);
}

void MidiChannelTrack::program(int channel, int prg)
{
  if (channel_ == NOCHANNEL)
    channelused[channel] = 1;
  if (channel_ == channel)
    write->program(channel, prg);
}

void MidiChannelTrack::control(int channel, int what, int value)
{
  if (channel_ == NOCHANNEL)
    channelused[channel] = 1;
  if (channel_ == channel)
    write->control(channel, what, value);
}

void MidiChannelTrack::noteon(int channel, int note, int vel)
{
  if (channel_ == NOCHANNEL)
    channelused[channel] = 1;

  if (channel_ == channel)
    write->noteon(channel, note, vel);
}

void MidiChannelTrack::noteoff(int channel, int note, int vel)
{
  if (channel_ == NOCHANNEL)
    channelused[channel] = 1;
  if (channel_ == channel)
    write->noteoff(channel, note, vel);
}

void MidiChannelTrack::pitchbend(int channel, int val)
{
  if (channel_ == NOCHANNEL)
    channelused[channel] = 1;
  if (channel_ == channel)
    write->pitchbend(channel, val);
}

void MidiChannelTrack::polyaftertouch(int channel, int note, int val)
{
  if (channel_ == NOCHANNEL)
    channelused[channel] = 1;
  if (channel_ == channel)
    write->polyaftertouch(channel, note, val);
}

void MidiChannelTrack::aftertouch(int channel, int val)
{
  if (channel_ == NOCHANNEL)
    channelused[channel] = 1;
  if (channel_ == channel)
    write->aftertouch(channel, val);
}

void MidiChannelTrack::songpos(unsigned pos)
{
  if (channel_ == NOCHANNEL)
    write->songpos(pos);
}

void MidiChannelTrack::songselect(unsigned char song)
{
  if (channel_ == NOCHANNEL)
    write->songselect(song);
}

void MidiChannelTrack::tunerequest()
{
  if (channel_ == NOCHANNEL)
    write->tunerequest();
}

void MidiChannelTrack::timingclock()
{
  if (channel_ == NOCHANNEL)
    write->timingclock();
}

void MidiChannelTrack::start()
{
  if (channel_ == NOCHANNEL)
    write->start();
}

void MidiChannelTrack::cont()
{
  if (channel_ == NOCHANNEL)
    write->cont();
}

void MidiChannelTrack::stop()
{
  if (channel_ == NOCHANNEL)
    write->stop();
}

void MidiChannelTrack::activesense()
{
  if (channel_ == NOCHANNEL)
    write->activesense();
}

void MidiChannelTrack::sysex(int syslen, unsigned char* sysdata)
{
  if (channel_ == NOCHANNEL)
    write->sysex(syslen, sysdata);
}

void MidiChannelTrack::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}

void midifmt()
{
  write = new MidiWrite(output);
  assert(write != 0);
  fprintf(stderr, "write tempo track\n");
  MidiChannelTrack head(input, NOCHANNEL);  head.run();

  for (int channel = 0; channel < 16; channel++)
  if (channelused[channel])
  {
    fprintf(stderr, "write track (channel %d)\n", channel+1);
    MidiChannelTrack track(input, channel);  track.run();
  }
  delete write;
  fprintf(stderr, "output written to %s\n", output);
}

void usage()
{
  fprintf(stderr, "usage: midi0to1 srcfile.mid dstfile.mid\n");
  fprintf(stderr, "split channels into separate tracks\n");
  fprintf(stderr, "This program only converts format 0 midi files.\n");
  exit(1);
}

int main(int argc, char** argv)
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert midi to same file\n");
    return 1;
  }

  midifmt();
  return 0;
}