// midi1to0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef __MSDOS__
#include <alloc.h>
#endif

static char* version = "midi1to0 v1.2 by Gnter Nagler (" __DATE__ ")";

class MidiCopyEvent;

typedef int int128[128];

struct MIDIPAR
{
  MIDIPAR();

  int128 ctrl;           // for all control numbers
  int prog;
  int128 poly;       // for all notes
  int aftertouch;
  int pitchbend;
};

MIDIPAR::MIDIPAR()
{
  for (int i = 0; i < 128; i++)
    ctrl[i] = poly[i] = -1;
  prog = -1;
  aftertouch = -1;
  pitchbend = -1;
}

char* input = 0;
char* output = 0;

int srcformat = 0;
int trackcount = 0;
long *trackend = 0;
long *trackpos = 0;
long *tracktime = 0;
MidiCopyEvent **trackread = 0;
unsigned long tracksum = 0, tracktotal = 0;

MidiWrite* write = 0;
int quiet = 0;
int exact = 0;

// For tracks that share a channel number we do careful merging by
// remembering all its parameters and set parameters before they will
// be applied

MIDIPAR param[16];
int curtrack;

class MidiScan : public MidiRead
{
public:
  MidiScan(char* name);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);


  virtual void error(const char* msg);

};

MidiScan::MidiScan(char* name) : MidiRead(name)
{
  options_ = OPTION_NOCONTROLS+OPTION_NOMETAEVENTS+
	     OPTION_NOEVENTS+OPTION_NOSYSEVENTS;
}

void MidiScan::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  if (version == 0)
  {
    if (!quiet)
      fprintf(stderr, "%s is already format 0\n", midiname_);
    exit(0);
  }
  if (version != 1)
  {
    fprintf(stderr, "%s is format %d. This program does not convert it to 0.\n", midiname_, version);
    exit(1);
  }
  srcformat = version;
  trackcount = tracks;
  assert(trackcount > 0);
  trackpos = (long*)malloc(trackcount * sizeof(long));
  trackend = (long*)malloc(trackcount * sizeof(long));
  tracktime = (long*)malloc(trackcount * sizeof(long));
  trackread = (MidiCopyEvent**)malloc(trackcount * sizeof(MidiCopyEvent*));
  assert(trackend != 0 && trackpos != 0 && tracktime != 0 && trackread != 0);
}

void MidiScan::track(int trackno, long length, int channel)
{
  trackend[trackno-1] = getcurpos() + length;
  trackpos[trackno-1] = getcurpos();
  tracktotal += length;
  tracktime[trackno-1] = 0; // current time at this track
}

void MidiScan::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}

class MidiCopyEvent : public MidiRead
{
public:
  MidiCopyEvent(char* name, FILE* f = 0);

  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void error(const char* msg);

  MIDIPAR param_;
};

MidiCopyEvent::MidiCopyEvent(char* name, FILE* f) : MidiRead(name, f)
{
  options_ = OPTION_NOCONTROLS+OPTION_NOMETAEVENTS+OPTION_NOSYSEVENTS;
}

void MidiCopyEvent::meta(int what, int len, unsigned char* data)
{
  if (what == meta_endtrack) // do not produce end track!
    return;
  write->meta(what, len, data);
}

void MidiCopyEvent::program(int channel, int prg)
{
  write->program(channel, prg);
  param[channel].prog = prg;
  param_.prog = prg;
}

void MidiCopyEvent::control(int channel, int what, int value)
{
  write->control(channel, what, value);
  param[channel].ctrl[what] = value;
  param_.ctrl[what] = value;
}

void MidiCopyEvent::noteon(int channel, int note, int vel)
{
  // look if parameters have changed
  if (exact)
  {
    for (int i = 0; i < 128; i++)
    {
      if (param[channel].ctrl[i] != param_.ctrl[i])
	control(channel, i, param_.ctrl[i]);
      if (param[channel].poly[i] != param_.poly[i])
	polyaftertouch(channel, i, param_.poly[i]);
    }
    if (param[channel].aftertouch != param_.aftertouch)
      aftertouch(channel, param_.aftertouch);
    if (param[channel].pitchbend != param_.pitchbend)
      pitchbend(channel, param_.pitchbend);
    if (param[channel].prog != param_.prog)
      program(channel, param_.prog);
  }
  write->noteon(channel, note, vel);
}

void MidiCopyEvent::noteoff(int channel, int note, int vel)
{
  write->noteoff(channel, note, vel);
}

void MidiCopyEvent::pitchbend(int channel, int val)
{
  write->pitchbend(channel, val);
  param[channel].pitchbend = val;
  param_.pitchbend = val;
}

void MidiCopyEvent::polyaftertouch(int channel, int note, int val)
{
  write->polyaftertouch(channel, note, val);
  param[channel].poly[note] = val;
  param_.poly[note] = val;
}

void MidiCopyEvent::aftertouch(int channel, int val)
{
  write->aftertouch(channel, val);
  param[channel].aftertouch = val;
  param_.aftertouch = val;
}

void MidiCopyEvent::songpos(unsigned pos)
{
  write->songpos(pos);
}

void MidiCopyEvent::songselect(unsigned char song)
{
  write->songselect(song);
}

void MidiCopyEvent::tunerequest()
{
  write->tunerequest();
}

void MidiCopyEvent::timingclock()
{
  write->timingclock();
}

void MidiCopyEvent::start()
{
  write->start();
}

void MidiCopyEvent::cont()
{
  write->cont();
}

void MidiCopyEvent::stop()
{
  write->stop();
}

void MidiCopyEvent::activesense()
{
  write->activesense();
}

void MidiCopyEvent::sysex(int syslen, unsigned char* sysdata)
{
  write->sysex(syslen, sysdata);
}

void MidiCopyEvent::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}


void midi1to0()
{

  MidiScan mainmidi(input);
  if (!mainmidi.getf())
  {
    perror(input);
    exit(1);
  }
  mainmidi.run();

  for (int trk = 0; trk < trackcount; trk++)
  {
    trackread[trk] = new MidiCopyEvent(input, mainmidi.getf()); // sharing file!
    if (!trackread[trk])
      continue; // out of memory: must ignore track

    // get time of first event
    trackread[trk]->seek(trackpos[trk]);
    tracktime[trk] = trackread[trk]->getdelta();
    tracksum += trackread[trk]->getcurpos() - trackpos[trk];
    trackpos[trk] = trackread[trk]->getcurpos();
  }

  write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return;
  }

  write->head(0, 1, mainmidi.unitsperbeat_); // midi header format 0, 1 track

  write->track();  // create single track of format 0

  unsigned long curtime = 0;
  for (;;)
  {
    int besttrk = -1;

    static int lastperc = -1;

    int perc;

    if (tracksum == 0)
      perc = 0;
    else
      perc = (int)((tracksum * 100) / tracktotal);

    if (perc != lastperc)
    {
      if (!quiet)
	fprintf(stderr, "%d%%\r", perc);
      lastperc = perc;
    }
    // find track that has lowest time > 0
    for (trk = 0; trk < trackcount; trk++)
    if (trackread[trk])
    {
      if (besttrk < 0 || tracktime[trk] < tracktime[besttrk])
      {
	besttrk = trk;
	if (tracktime[trk] == curtime)
	  break;
      }
    }
    if (besttrk < 0)  // no more tracks with events
      break;

    tracksum -= trackpos[besttrk];
    write->time(tracktime[besttrk] - curtime);
    curtime = tracktime[besttrk];
    // copy event:
    curtrack = besttrk;
    trackread[besttrk]->seek(trackpos[besttrk]);
    trackread[besttrk]->runevent(trackend[besttrk]);

    if (trackread[besttrk]->getcurpos() >= trackend[besttrk])
    {
      delete trackread[besttrk];
      trackread[besttrk] = 0;
    }
    else
    {
      tracktime[besttrk] += trackread[besttrk]->getdelta();
      trackpos[besttrk] = trackread[besttrk]->getcurpos();
      if (trackpos[besttrk] + 3 <= trackend[besttrk] &&
	  memcmp(trackread[besttrk]->get(3), "\xff\x2f", 3) == 0) // meta event end track
      {
	delete trackread[besttrk];
	trackread[besttrk] = 0;
      }
    }
    tracksum += trackpos[besttrk];
  }

  write->endtrack(); // close the track

  delete write;  // close midi file
  if (!quiet)
    fprintf(stderr, "output written to %s\n", output);

  free(trackend);
  free(trackpos);
  free(tracktime);
  free(trackread);
}

void usage()
{
  fprintf(stderr, "usage: midi1to0 [-quiet] [-exact] srcfile.mid dstfile.mid\n");
  fprintf(stderr, "-quiet\tbe quiet, don't tell percentage\n");
  fprintf(stderr, "-exact\tsolve merging tracks with conflicting channel by \n");
  fprintf(stderr, "\trefreshing all track parameters before playing notes\n");
  fprintf(stderr, "midi1to0 merges channels into one track\n");
  fprintf(stderr, "This program only converts format 1 midi files.\n");
  exit(1);
}

int main(int argc, char** argv)
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (strncmp(*argv, "-quiet", 2) == 0)
    {
      quiet = 1;
      argc--; argv++;
      continue;
    }
    if (strncmp(*argv, "-exact", 2) == 0)
    {
      exact = 1;
      argc--; argv++;
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert midi to same file\n");
    return 1;
  }

  midi1to0();
  return 0;
}
