// midi2gm written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static char* version = "midi2gm v1.5 by Gnter Nagler (" __DATE__ ")";

#ifdef __MSDOS__
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#else
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#endif

// minimal length to initialize track
#define MINTRACKLEN 36

char* input = 0;
char* output = 0;

MidiWrite* write = 0;
int i3 = 0;
int iskorgi2i3 = 0;
int changed = 0;
int quiet = 0;
int notext = 0;
int nosys = 0;
int nobend[16] = {   // no pitch bend events on this channel
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0
};
int notbank = 0;
int playchannel = 0; // forget only notes at this channel
int nonote[16] = {   // no noteon/noteoff events on this channel
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0
};

int mapchannel[16] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
int xg_reset = 0;
int gs_reset = 0;
int gm_reset = 0;

// korg i2/i3 parameter initialization
unsigned char def_reverb[16] = { 28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28 };
unsigned char def_chorus[16] = { 28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28 };

class Midi2GM : public MidiCopy
{
public:
  Midi2GM(char* name);

  virtual void track(int trackno, long length, int channel);

  virtual void text(int what, int len, char* whattext, unsigned char* txt);
  virtual void meta(int what, int len, unsigned char* data);
  virtual void program(int prg, int channel);
  virtual void highbank(int channel, int val);
  virtual void lowbank(int channel, int val);
  virtual void pitchbend(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void sysex(int syslen, unsigned char* sysdata);

  virtual void smpteofs(int hour, int min, int sec, int frame, int fracframe);

  virtual void error(const char* msg);
  virtual void warning(const char* msg);

  virtual void percent(int perc);

  void i3_initchannel(int channel);

};

Midi2GM::Midi2GM(char* name) : MidiCopy(name)
{
}

void Midi2GM::i3_initchannel(int channel)
{
    if (channel < 0)
      return;
    if (mapchannel_[channel] == gm_drumchannel)
    {
      // select bank Drum1 gm-drums
      MidiCopy::highbank(channel, 0x3e);
      MidiCopy::lowbank(channel, 0);
      MidiCopy::program(channel, 0); // Dr1
      MidiCopy::volume(channel, 127);
    }
    else
    {
      // select bank AB
      MidiCopy::highbank(channel, 0x38);
      MidiCopy::lowbank(channel, 0);
      MidiCopy::program(channel, 0); // gm-piano
      MidiCopy::volume(channel, 127);
    }
}

void Midi2GM::track(int trackno, long length, int channel)
{
  MidiCopy::track(trackno, length, channel);
  if (trackno == 1)
  {
    if (xg_reset)
      MidiCopy::xgreset();
    if (gs_reset)
      MidiCopy::gsreset();
    if (gm_reset)
      MidiCopy::gmreset();

    MidiCopy::tempo(500000L);  // default tempo
    MidiCopy::tact(4,4,24,8);  // default tact 4/4
    changed = 1;
  }
  if (i3 && length >= MINTRACKLEN)  // initialize keyboard program and bank
  {
    changed = 1;

     if (channel == MULTICHANNEL)
     {
       for (int channel = 0; channel < 16; channel++)
	 i3_initchannel(channel);
     }
     else
	 i3_initchannel(channel);
  }
  if (iskorgi2i3)
  {
    // initialize reverb & chorus

     if (channel == MULTICHANNEL)
     {
       fprintf(stderr, "korg i2/i3 reverb and chorus initialization\n");
       for (int channel = 0; channel < 16; channel++)
       {
	 MidiCopy::reverb(channel, def_reverb[channel]);
	 MidiCopy::chorus(channel, def_chorus[channel]);
       }
     }
     else
     {
       fprintf(stderr, "korg i2/i3 reverb and chorus initialization in channel %d\n", channel+1);
       MidiCopy::reverb(channel, def_reverb[channel]);
       MidiCopy::chorus(channel, def_chorus[channel]);
     }
  }
}

void Midi2GM::text(int what, int len, char* whattext, unsigned char* txt)
{
  if (!notext)
    MidiCopy::text(what, len, whattext, txt);
  if (what == meta_trackname && version_ == VERSION_SINGLECHANNEL &&
     VALIDCHANNEL(getchannel()) && getchannel() != gm_drumchannel &&
     mapchannel_[getchannel()] == getchannel())
  {  // check if this track is a drum track inspite of not having gm drum channel
  char instr[256];

     strncpy(instr, txt, len);
     instr[len] = '\0';

#ifdef __MSDOS__
     strupr(instr);
#else
     // upper case instrument name
     for (char* instrp = instr; *instrp; instrp++)
       *instrp = toupper(*instrp);
#endif
     if ((strstr(instr, "DRUM") != 0 && strstr(instr, "STEEL") == 0) ||
	 strstr(instr, "PERCUSSION") != 0 ||
	 strstr(instr, "HIHAT") != 0 ||
	 strstr(instr, "SNARE") != 0 ||
	 strstr(instr, "KICK") != 0 ||
	 strstr(instr, "CRASH") != 0 ||
	 strstr(instr, "TAMBOURINE") != 0 ||
	 strstr(instr, "CLICK") != 0 ||
	 strstr(instr, "HI-HAT") != 0)
     {
       if (!quiet)
	 fprintf(stderr, "map drum channel %d to channel 10\n", SAYCHANNEL(getchannel()));
       mapchannel_[getchannel()] = gm_drumchannel;
       if (i3)
	 i3_initchannel(gm_drumchannel);
     }
  }
}

void Midi2GM::meta(int what, int len, unsigned char* data)
{
  if (what == 0x7F && len == 299 && data[0] == 'B' && data[1] == '9' && data[2] == 1)
  {
    fprintf(stderr, "korg i2/i3 meta event found and ignored\n");

    // korg i2/i3 song parameter dump
    iskorgi2i3 = 1;
    unsigned char* trackparam = data + 96;

    const level2ctrl[16] = { 0, 14, 28, 42, 56, 71, 85, 99, 113, 127, 127, 127, 127, 127, 127, 127 };

    for (int ch = 0; ch < 16; ch++, trackparam += 12)
    {
      def_reverb[ch] = level2ctrl[(trackparam[5] & 0x0f)];
      def_chorus[ch] = level2ctrl[(trackparam[5] >> 4) & 0x0f];
    }
    changed = 1;
  }
  else
    MidiCopy::meta(what, len, data);
}

void Midi2GM::program(int channel, int prg)
{
  if (mapchannel_[channel] != gm_drumchannel)
    MidiCopy::program(channel, prg);
  else
  {
    if (!quiet)
       fprintf(stderr, "program %d in channel %d ignored\n", prg, SAYCHANNEL(mapchannel_[channel]));
    changed++;
  }
}

void Midi2GM::pitchbend(int channel, int val)
{
  if (nobend[channel])
    return;
  MidiCopy::pitchbend(channel, val);
}

void Midi2GM::highbank(int channel, int val)
{
  if (notbank)
  {
    MidiCopy::highbank(channel, val);
    return;
  }
  if (!quiet)
    fprintf(stderr, "hbank %d in channel %d ignored\n", val, SAYCHANNEL(mapchannel_[channel]));
  changed++;
}

void Midi2GM::lowbank(int channel, int val)
{
  if (notbank)
  {
    MidiCopy::lowbank(channel, val);
    return;
  }
  if (!quiet)
    fprintf(stderr, "lowbank %d in channel %d ignored\n", val, SAYCHANNEL(mapchannel_[channel]));
  changed++;
}

void Midi2GM::noteon(int channel, int note, int vel)
{
  if (nonote[channel])
  {
    if (changed==0)
      changed++;
    return;
  }
  MidiCopy::noteon(channel, note, vel);
}

void Midi2GM::noteoff(int channel, int note, int vel)
{
  if (nonote[channel])
    return;
  MidiCopy::noteoff(channel, note, vel);
}

void Midi2GM::sysex(int syslen, unsigned char* sysdata)
{
  if (nosys)
    return;
  MidiCopy::sysex(syslen, sysdata);
}

void Midi2GM::smpteofs(int , int , int , int , int )
{
  // ignore
}

void Midi2GM::percent(int perc)
{
  if (!quiet)
    fprintf(stderr, "%-3d%%\r", perc);
}

void Midi2GM::error(const char* msg)
{
  printf("Error: %s\n", msg);
  exit(1);
}

void Midi2GM::warning(const char* msg)
{
  printf("Warning: %s\n", msg);
}

void usage()
{
  fprintf(stderr, "Midi2GM removes bank select commands and program changes in drum channel 10\n");
  fprintf(stderr, "usage: Midi2GM [-i3][-q][-not][-nos][-non #][-map #[-#] #][-notb][-xg][-gs][-gm] file.mid korg.mid\n");
  fprintf(stderr, "-version\tget program version\n");
  fprintf(stderr, "-i3\t(for korg i2/i3 only) initializes banks, programs, volumes\n");
  fprintf(stderr, "-q\tbe quiet\n");
  fprintf(stderr, "-notext\tignore all textual information (tracknames, lyrics,...)\n");
  fprintf(stderr, "-nosys\tignore all sysex events\n");
  fprintf(stderr, "-nonote #\tignore all note events on list of channels (1-16)\n");
  fprintf(stderr, "-nobend #\tNo pitch bend effects on this channel (1-16)\n");
  fprintf(stderr, "-mapchannel c[-d] e\tmap events of channels c..d to channel e (1..16)\n");
  fprintf(stderr, "-notbank\tbank change commands are not removed\n");
  fprintf(stderr, "-xg\tadd XG reset command\n");
  fprintf(stderr, "-gs\tadd GS reset command\n");
  fprintf(stderr, "-gm\tadd GM reset command\n");
  exit(1);
}

void copyfile(char* input, char* output)
{
FILE* i = fopen(input, READ_BINARY);

  if (!i)
  {
    perror(input);
    return;
  }
FILE* o = fopen(output, WRITE_BINARY);
  if (!o)
  {
    fclose(i);
    perror(output);
    return;
  }
  static char buf[4096];
  int n;
  long pos = 0;

  while (fread(buf, sizeof(buf), 1, i) == 1)
  {
    fwrite(buf, sizeof(buf), 1, o);
    pos += sizeof(buf);
  }
  fseek(i, pos, SEEK_SET);
  fseek(o, pos, SEEK_SET);
  while ((n = fread(buf, 1, sizeof(buf), i)) > 0)
    fwrite(buf, n, 1, o);

  fclose(i);
  fclose(o);
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (argc > 0 && strcmp(*argv, "-i3") == 0)
    {
      i3 = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strcmp(*argv, "-q") == 0)
    {
      quiet = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-notbank", 5) == 0)
    {
      notbank = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-notext", 4) == 0)
    {
      notext = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-nosys", 4) == 0)
    {
      nosys = 1;
      argc--; argv++;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-nonote", 4) == 0)
    {
      argc--; argv++;
      if (!argc)
	fprintf(stderr, "option -nonote expects a channel number 1-16\n");
      while (argc > 0 && isdigit(**argv))
      {
	int fromchannel, tochannel;

	if (sscanf(*argv, "%u-%u", &fromchannel, &tochannel) == 2)
	{
	  for (int channel = fromchannel; channel <= tochannel; channel++)
	  if (channel >= 1 && channel <= 16)
	    nonote[channel-1] = 1;
	}
	else
	{
	  int channel = atoi(*argv);
	  if (channel <= 0 || channel > 16)
	    fprintf(stderr, "channel number %d out of range 1-16\n", channel);
	  else
	    nonote[channel-1] = 1;
	}
	argc--; argv++;
      }
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-nobend", 4) == 0)
    {
      argc--; argv++;
      if (!argc)
	fprintf(stderr, "option -nobend expects a channel number 1-16\n");
      while (argc > 0 && isdigit(**argv))
      {
	int fromchannel, tochannel;

	if (sscanf(*argv, "%u-%u", &fromchannel, &tochannel) == 2)
	{
	  for (int channel = fromchannel; channel <= tochannel; channel++)
	  if (channel >= 1 && channel <= 16)
	    nobend[channel-1] = 1;
	}
	else
	{
	  int channel = atoi(*argv);
	  if (channel <= 0 || channel > 16)
	    fprintf(stderr, "channel number %d out of range 1-16\n", channel);
	  else
	    nobend[channel-1] = 1;
	}
	argc--; argv++;
      }
      continue;
    }
    if (strncmp(*argv, "-mapchannel", 2) == 0)
    {
    int firstchannel, secondchannel, destchannel;

      argc--; argv++;
      if (argc > 0 && sscanf(*argv, "%u-%u", &firstchannel, &secondchannel) == 2)
      {
	if (firstchannel <= 0 || secondchannel > 16)
	{
	  fprintf(stderr, "channel number %s out of range 1-16\n", *argv);
	  firstchannel = secondchannel = -1;
	}
	else if (firstchannel > secondchannel)
	{
	  fprintf(stderr, "%s: first channel number must be less than second one\n", *argv);
	  firstchannel = secondchannel = -1;
	}
	argc--; argv++;
      }
      else if (argc > 0 && isdigit(**argv))
      {
	firstchannel = atoi(*argv);
	if (firstchannel <= 0 || firstchannel > 16)
	{
	  fprintf(stderr, "channel number %s out of range 1-16\n", *argv);
	  firstchannel = -1;
	}
	secondchannel = firstchannel;
	argc--; argv++;
      }
      else
      {
	fprintf(stderr, "option -mapchannel expects two channel numbers 1-16\n");
	usage();
	break;
      }
      if (argc > 0 && isdigit(**argv))
      {
	destchannel = atoi(*argv);
	argc--; argv++;
      }
      else
      {
	fprintf(stderr, "option -mapchannel expects two channel numbers 1-16\n");
	usage();
	break;
      }
      for (int c = firstchannel; c <= secondchannel; c++)
      if (c != destchannel)
      {
	mapchannel[c-1] = destchannel-1;
	changed++;
      }
      continue;
    }
    if (strcmp(*argv, "-xg") == 0)
    {
      xg_reset = 1;
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-gs") == 0)
    {
      gs_reset = 1;
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-gm") == 0)
    {
      gm_reset = 1;
      argc--; argv++; continue;
    }

    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert midi to same file\n");
    return 1;
  }
  Midi2GM midi(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }
  write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return 1;
  }
  if (!write->getf())
  {
    perror(output);
    return 1;
  }
  argc--; argv++;

  for (int c = 0; c < 16; c++)
    midi.mapchannel(c, mapchannel[c]);

  midi.setoutput(write);

  midi.options_ = OPTION_NOSYSEVENTS;
  if (!midi.run())
    fprintf(stderr, "%s: midi read error at %04lX\n", input, midi.getpos());

  delete write;

  if (!changed)
  {
    copyfile(input, output);
    if (!quiet)
       fprintf(stderr, "%s: not changed. Now you have a nice copy.\n", input);
  }
  return 0;
}
