// midi2hex v1.0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#ifdef __MSDOS__
#include <dos.h>
#else
#endif

MidiRead* midi = 0;
char* midiname = 0;
int format = 0;
int tracks = 0;
int resolution = 0;

void usage()
{
  fprintf(stderr, "usage: midi2hex file.mid [> file.txt]\n");
  exit(1);
}

void hex(unsigned char* s, int len, int space = 1)
{
  assert(s != 0);
  while (len > 0)
  {
    if (space)
      putchar(' ');
    printf("%02X", *s++);
    len--;
  }
}

void hex(long len)
{
  while (len > 0)
  {
    hex(midi->get(1), 1);
    len--;
  }
}

int head2hex(long len)
{
  if (len != 6)
  {
    fprintf(stderr, "invalid midi header\n");
    return 0;
  }
  long pos = midi->getcurpos();
  format = midi->getword();
  tracks = midi->getword();
  resolution = midi->getword();

  midi->seek(pos);
  printf("  ");
  hex(midi->get(2), 2, 0);
  printf(" ");
  hex(midi->get(2), 2, 0);
  printf(" ");
  hex(midi->get(2), 2, 0);
  printf("  // format, trackcount, resolution\n");
  return 1;
}

int deltalen = 0;

long delta2hex()
{
unsigned char* s;

  deltalen = 0;
  long pos = midi->getcurpos();
  while (deltalen < 4)
  {
    s = midi->get(1);
    deltalen++;
    if ((*s & 0x80) == 0)
      break;
  }
  midi->seek(pos);
  long value = midi->getdelta();
  midi->seek(pos);
  hex(midi->get(deltalen), deltalen, 0);
  return value;
}

int track2hex(long len)
{
int lastcode = -1;
unsigned char* s, midicode;

  while (len > 0)
  {
    // time code
    long pos = midi->getcurpos();

    printf("%04lX: ", pos);
    delta2hex();
    putchar('\t');

    if (deltalen > len)
    {
	fprintf(stderr, "unexpected end of track\n");
	return 0;
    }
    len -= deltalen;
    pos += deltalen;
    s = midi->get(1);
    if (!s || s[0] >= 0x80 || lastcode < 0)
    {
      midicode = *s;
      printf("%02X", midicode);
      len--;
    }
    else
    {
      printf("  ");
      midicode = lastcode;
      midi->seek(pos);
    }
    if (midicode < 0x80)
    {
      fprintf(stderr, "invalid midi command found: %02X\n", midicode);
      return 0;
    }
    putchar(' ');
    if ((midicode & 0xF0) == 0xF0)
    {
      switch(midicode)
      {
      case 0xF0: // sysex ended by F7
      case 0xF7: // old variant of sysex packages
	int syslen = midi->scansysevent(len);
	if (syslen == 0)
	{
	    fprintf(stderr, "end of sysex command (F7) missing\n");
	    break;
	}
	while (syslen-- > 0)
	{
	  s = midi->get(1);  len--;
	  hex(s, 1);
	}
	printf(" // sysex");
	break;
      case 0xf2:
	hex(2);
	len -= 2;
	printf(" // song pos");
	break;
      case 0xf3:
	hex(1);
	len -= 2;
	printf(" // song select");
	break;
      case 0xf6:
	printf(" // tune request");
	break;
      case 0xf8:
	printf(" // clock");
	break;
      case 0xfa:
	printf(" // start");
	break;
      case 0xfb:
	printf(" // continue");
	break;
      case 0xfc:
	printf(" // stop");
	break;
      case 0xfe:
	printf(" // active sense");
	break;
      case 0xFF:  // meta
	// what kind of meta
	s = midi->get(1);
	hex(s, 1); len--;
	printf(" ");
	// what length
	long metalen = delta2hex();
	len -= deltalen;
	len -= metalen;
	printf(" ");
	hex(metalen);
	switch(*s)
	{
	case 0:
	  printf(" // sequence number");
	  break;
	case meta_text:
	  printf(" // text");
	  break;
	case meta_copyright:
	  printf(" // copyright");
	  break;
	case meta_trackname:
	  printf(" // trackname");
	  break;
	case meta_instrument:
	  printf(" // instrument");
	  break;
	case meta_lyric:
	  printf(" // lyric");
	  break;
	case meta_marker:
	  printf(" // marker");
	  break;
	case meta_cuepoint:
	  printf(" // cue point");
	  break;
	case 8:
	case 9:
	case 10:
	case 11:
	case 12:
	case 13:
	case 14:
	case 15:
	  printf(" // meta text");
	  break;
	case 0x2F:
	  printf(" // end of track");
	  break;
	case 0x51:
	  printf(" // tempo");
	  break;
	case 0x54:
	  printf(" // smpte timecode");
	  break;
	case 0x58:
	  printf(" // tact");
	  break;
	case 0x59:
	  printf(" // key");
	  break;
	default:
	  printf(" // meta");
	  break;
	}
	break;
      default:
	fprintf(stderr, "invalid midi command %02X\n", midicode);
	break;
      }
    }
    else
    {
    static char* comment[7] =
     {
       "note off",  // 0x8?
       "note on",   // 0x9?
       "poly aftertouch", // 0xA?
       "control",   // 0xB?
       "program",   // 0xC?
       "aftertouch", // 0xD?
       "pitch bend", // 0xE?
     };

      if (format == 1 && lastcode >= 0x80)
      {
	if ((midicode & 15) != (lastcode & 15))
	  fprintf(stderr, "more than one channels (%d,%d) used in format 1 track!\n",
	    midicode & 15, lastcode & 15);
      }
      switch(midicode & 0xF0)
      {
      case 0x80:
      case 0x90:
      case 0xA0:
      case 0xE0:
	s = midi->get(2);
	hex(s, 2);
	len -= 2;
	if (s[1] == 0) // velocity == 0 => note off
	  printf(" // %s", comment[0]);
	else
	  printf(" // %s", comment[ (midicode >> 4) - 8 ]);
	break;
      case 0xB0:
	s = midi->get(2);
	hex(s, 2);
	len -= 2;
	switch(*s)
	{
	case ctrl_highbank: printf(" // high bank"); break;
	case ctrl_wheel: printf(" // wheel"); break;
	case ctrl_breath: printf(" // breath"); break;
	case ctrl_foot: printf(" // foot"); break;
	case ctrl_portamentotime: printf(" // portamentotime"); break;
	case ctrl_data: printf(" // data"); break;
	case ctrl_volume: printf(" // volume"); break;
	case ctrl_balance: printf(" // balance"); break;
	case ctrl_expression: printf(" // expression"); break;
	case ctrl_lowbank: printf(" // lowbank"); break;
	case ctrl_hold: printf(" // hold"); break;
	case ctrl_reverb: printf(" // reverb"); break;
	case ctrl_chorus: printf(" // chorus"); break;
	case ctrl_datainc: printf(" // datainc"); break;
	case ctrl_datadec: printf(" // datadec"); break;
	default:
	  printf(" // control");
	  break;
	}
	break;
      case 0xC0:
      case 0xD0:
	hex(1);
	len -= 1;
	printf(" // %s", comment[ (midicode >> 4) - 8 ]);
	break;
      }
      lastcode = midicode;
    }
    putchar('\n');
    if (len < 0)
    {
      fprintf(stderr, "unexpected end of track\n");
      return 0;
    }
  }
  return 1;
}

int midi2hex()
{
  unsigned char* s;

  s = midi->get(4);
  if (strncmp(s, "MThd", 4) != 0)
  {
    fprintf(stderr, "%s is not a midi file\n", midiname);
    return 0;
  }
  printf("// %s\n", midiname);
  long pos = midi->getcurpos();
  s = midi->get(4);
  printf("MThd %02X%02X%02X%02X\n", s[0], s[1], s[2], s[3]);
  midi->seek(pos);
  long taglen = midi->getlong();

  long tagpos = midi->getcurpos();
  if (!head2hex(taglen))
    return 0;

  midi->seek(tagpos + taglen);

  for (int i = 0; i < tracks; i++)
  {
    s = midi->get(4);
    if (strncmp(s,"MTrk", 4) != 0)
    {
      fprintf(stderr, "invalid tag in midi\n");
      return 0;
    }
    pos = midi->getcurpos();
    s = midi->get(4);
    printf("\nMTrk %02X%02X%02X%02X\n", s[0], s[1], s[2], s[3]);
    midi->seek(pos);
    taglen = midi->getlong();
    if (!track2hex(taglen))
      break;
  }
  return 1;
}

int main(int argc, char**argv)
{
#ifdef __MSDOS__
  setcbrk(1);
#endif
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc == 0)
    usage();

  midiname = *argv;
  midi = new MidiRead(midiname); // use reader directly without parser
  if (!midi)
  {
    fprintf(stderr, "out of memory\n");
    return 1;
  }
  if (!midi->getf())
  {
    perror(midiname);
    return 1;
  }
  if (!midi2hex())
  {
    delete midi;
    return 1;
  }
  delete midi;
  return 0;
}

