// mididrum written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef __MSDOS__
#include <alloc.h>
#endif

static char* version = "mididrum v1.5 by Gnter Nagler (" __DATE__ ")";

int isdrum[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 };
#define MAXDRUMS  128
int drumused[MAXDRUMS];
#define MAXPROGRAMS 128
int programused[16][MAXPROGRAMS];
char noteplaying[16][128];

class MidiCopyEvent;

char* input = 0;
char* output = 0;

int splitprograms = 0;
int splitdrums = 0;

int trackcount = 0, trackno = -1;
long *trackend = 0;
long *trackpos = 0, *firsttrackpos = 0;
long *tracktime = 0, *firsttracktime = 0;
int  *trackchannel = 0;
int  *trackprograms = 0;

MidiCopyEvent **trackread = 0;
unsigned long tracksum = 0, tracktotal = 0, firsttracksum = 0;

typedef int int128[128];
int128 *trackctrl;
int128 ctrl; // current values
int onlynote = -1; // < 0 if any note, otherwise only notes that match
int onlyprogram = -1;
int onlychannel = -1;
int drumchannel = 9;  // 0..15

MidiWrite* write = 0;

class MidiScan : public MidiRead
{
public:
  MidiScan(char* name);

  virtual void head(unsigned version, unsigned tracks, unsigned unitsperbeat);
  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);

  virtual void program(int channel, int prg);

  virtual void error(const char* msg);

  int curprogram_[16];
};

MidiScan::MidiScan(char* name) : MidiRead(name)
{
  options_ = OPTION_NOCONTROLS+OPTION_NOMETAEVENTS+OPTION_NOSYSEVENTS;
}

void MidiScan::head(unsigned version, unsigned tracks, unsigned unitsperbeat)
{
  if (version == VERSION_MULTICHANNEL)
  {
    splitdrums = splitprograms = 1; // must split multichannel
  }
  else if (version != VERSION_SINGLECHANNEL)
  {
    fprintf(stderr, "%s is format %d. This program only accepts format 0 or 1.\n",
        midiname_, version);
    exit(1);
  }
  trackcount = tracks;
  assert(trackcount > 0);
  trackpos = (long*)malloc(trackcount * sizeof(long));
  firsttrackpos = (long*)malloc(trackcount * sizeof(long));
  trackend = (long*)malloc(trackcount * sizeof(long));
  tracktime = (long*)malloc(trackcount * sizeof(long));
  firsttracktime = (long*)malloc(trackcount * sizeof(long));
  trackchannel = (int*)malloc(trackcount * sizeof(int));
  trackprograms = (int*)malloc(trackcount * sizeof(int));
  trackctrl = (int128*)malloc(trackcount * sizeof(int128));
  trackread = (MidiCopyEvent**)malloc(trackcount * sizeof(MidiCopyEvent*));
  assert(trackend != 0 && trackpos != 0 && tracktime != 0 &&
	 trackread != 0 && trackchannel != 0 && trackctrl != 0 &&
	 firsttrackpos != 0 && firsttracktime != 0 && trackprograms != 0);
  for (int i = 0; i < MAXDRUMS; i++)
    drumused[i] = 0;
}

void MidiScan::track(int trackno, long length, int channel)
{
  trackend[trackno-1] = getcurpos() + length;
  trackpos[trackno-1] = getcurpos();
  tracktotal += length;
  tracktime[trackno-1] = 0; // current time at this track
  trackchannel[trackno-1] = channel; // current time at this track
  trackprograms[trackno-1] = 0;
  if (!splitprograms && (channel < 0 || !isdrum[channel]))
    options_ |= OPTION_NOEVENTS;
  else
    options_ &= ~OPTION_NOEVENTS;
  for (int ch= 0; ch < 16; ch++)
  for (int i = 0; i < MAXPROGRAMS; i++)
    programused[ch][i] = 0;
  for (ch = 0; ch < 16; ch++)
    curprogram_[ch] = -1;
}

void MidiScan::endtrack(int trackno)
{
  int cnt  = 0;
  for (int ch = 0; ch < 16; ch++)
  for (int i = 0; i < MAXPROGRAMS; i++)
  if (programused[i])
    cnt++;
  trackprograms[trackno-1] = cnt;
}

void MidiScan::noteon(int channel, int note, int vel)
{
  // drum note!
  if (channel == drumchannel)
  {
    drumused[note] = 1;
    return;
  }
  if (curprogram_[channel] < 0)
    curprogram_[channel] = 0;
  programused[channel][curprogram_[channel]] = 1;
}

void MidiScan::noteoff(int channel, int note, int vel)
{
  if (channel == drumchannel)
  {
    drumused[note] = 1;
    return;
  }
  if (curprogram_[channel] < 0)
    curprogram_[channel] = 0;
  programused[channel][curprogram_[channel]] = 1;
}

void MidiScan::program(int channel, int program)
{
  if (channel == drumchannel)
    return;
  curprogram_[channel] = program;
}

void MidiScan::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}

void mutenotes()
{
  for (int i = 0; i  < 128; i++)
  for (int c = 0; c < 16; c++)
    noteplaying[c][i] = 0;
}

class MidiCopyEvent : public MidiRead
{
public:
  MidiCopyEvent(char* name, FILE* f = 0);

  virtual void track(int trackno, long length, int channel);
  virtual void endtrack(int trackno);

  virtual void time(unsigned long ticks);

  virtual void program(int prg, int channel);
  virtual void control(int channel, int what, int value);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void pitchbend(int channel, int val);
  virtual void polyaftertouch(int channel, int note, int val);
  virtual void aftertouch(int channel, int val);
  virtual void songpos(unsigned pos);
  virtual void songselect(unsigned char song);
  virtual void tunerequest();
  virtual void timingclock();
  virtual void start();
  virtual void cont();
  virtual void stop();
  virtual void activesense();
  virtual void sysex(int syslen, unsigned char* sysdata);
    virtual void xgreset();
    virtual void gmreset();
    virtual void gsreset();
    virtual void gsexit();

  virtual void meta(int what, int len, unsigned char* data);
  virtual void text(int what, int len, char* whattext, unsigned char* txt);
  virtual void end();
  virtual void prefixchannel(unsigned char channel);
  virtual void prefixport(unsigned char port);
  virtual void smpteofs(int hour, int min, int sec, int frame, int fracframe);
  virtual void tact(int nom, int denom, int unitsperbeat, int notes32perbeat);
  virtual void tempo(unsigned long microsecperbeat);
  virtual void seqnumber(unsigned int seqno);
  virtual void key(int signature, int isminor); // C=0, -7=7flats +7=7sharps
  virtual void error(const char* msg);

  int curprogram_[16];
};

MidiCopyEvent::MidiCopyEvent(char* name, FILE* f) : MidiRead(name, f)
{
  options_ = OPTION_NOCONTROLS;
  for (int ch = 0; ch < 16; ch++)
    curprogram_[ch] = -1;
}

void MidiCopyEvent::track(int trackno, long length, int channel)
{
  for (int ch = 0; ch < 16; ch++)
    curprogram_[ch] = -1;
}

void MidiCopyEvent::endtrack(int trackno)
{
}

void MidiCopyEvent::time(unsigned long ticks)
{
  write->time(ticks);
}

void MidiCopyEvent::meta(int what, int len, unsigned char* data)
{
  write->meta(what, len, data);
}

void MidiCopyEvent::program(int channel, int prg)
{
  if (onlychannel >= 0 && onlychannel != channel)
    return;
  curprogram_[channel] = prg;
  if (onlynote >= 0)
    channel = drumchannel;
  if (onlyprogram < 0 || onlyprogram == prg)
    write->program(channel, prg);
}

void MidiCopyEvent::control(int channel, int what, int value)
{
  if (onlychannel >= 0 && onlychannel != channel)
    return;
  if (onlynote >= 0)
  {
    channel = drumchannel;
    assert(trackno >= 0 && trackno < trackcount);
    trackctrl[trackno][what] = value;
    ctrl[what] = value;
  }
  write->control(channel, what, value);
}

void MidiCopyEvent::noteon(int channel, int note, int vel)
{
  if (onlychannel >= 0 && onlychannel != channel)
    return;
  if (onlynote >= 0)
    channel = drumchannel;
  if (curprogram_[channel] < 0)
    curprogram_[channel] = 0;
  int certainnote = (onlynote < 0 || note == onlynote);
  int certainprog = (onlyprogram < 0 || onlyprogram == curprogram_[channel]);
  if (certainnote && certainprog)
  {
    write->noteon(channel, note, vel);
    noteplaying[channel][note] = 1;
  }
}

void MidiCopyEvent::noteoff(int channel, int note, int vel)
{
  if (onlychannel >= 0 && onlychannel != channel)
    return;
  if (onlynote >= 0)
    channel = drumchannel;
  int certainnote = (onlynote < 0 || note == onlynote);
  int certainprog = (onlyprogram < 0 || onlyprogram == curprogram_[channel]);
  if (certainnote && (certainprog || noteplaying[channel][note]))
  {
    write->noteoff(channel, note, vel);
    noteplaying[channel][note] = 0;
  }
}

void MidiCopyEvent::pitchbend(int channel, int val)
{
  if (onlychannel >= 0 && onlychannel != channel)
    return;
  if (onlynote >= 0)
    channel = drumchannel;
  write->pitchbend(channel, val);
}

void MidiCopyEvent::polyaftertouch(int channel, int note, int val)
{
  if (onlynote >= 0)
    channel = drumchannel;
  write->polyaftertouch(channel, note, val);
}

void MidiCopyEvent::aftertouch(int channel, int val)
{
  if (onlychannel >= 0 && onlychannel != channel)
    return;
  if (onlynote >= 0)
    channel = drumchannel;
  write->aftertouch(channel, val);
}

void MidiCopyEvent::songpos(unsigned pos)
{
  write->songpos(pos);
}

void MidiCopyEvent::songselect(unsigned char song)
{
  write->songselect(song);
}

void MidiCopyEvent::tunerequest()
{
  write->tunerequest();
}

void MidiCopyEvent::timingclock()
{
  write->timingclock();
}

void MidiCopyEvent::start()
{
  write->start();
}

void MidiCopyEvent::cont()
{
  write->cont();
}

void MidiCopyEvent::stop()
{
  write->stop();
}

void MidiCopyEvent::activesense()
{
  write->activesense();
}

void MidiCopyEvent::sysex(int syslen, unsigned char* sysdata)
{
  write->sysex(syslen, sysdata);
}

void MidiCopyEvent::xgreset()
{
  if (write->trackcount() <= 1)
    write->xgreset();
}

void MidiCopyEvent::gmreset()
{
  if (write->trackcount() <= 1)
    write->gmreset();
}

void MidiCopyEvent::gsreset()
{
  if (write->trackcount() <= 1)
    write->gsreset();
}

void MidiCopyEvent::gsexit()
{
  if (write->trackcount() <= 1)
    write->gsexit();
}

void MidiCopyEvent::seqnumber(unsigned int seqno)
{
  write->seqnumber(seqno);
}

void MidiCopyEvent::smpteofs(int hour, int min, int sec, int frame, int fracframe)
{
  write->smpteofs(hour, min, sec, frame, fracframe);
}

void MidiCopyEvent::key(int signature, int isminor)
{
  write->key(signature, isminor);
}

void MidiCopyEvent::tact(int nom, int denom, int unitsperbeat, int notes32perbeat)
{
  write->tact(nom, denom, unitsperbeat, notes32perbeat);
}

void MidiCopyEvent::tempo(unsigned long microsecperbeat)
{
  write->tempo(microsecperbeat);
}

void MidiCopyEvent::end()
{
  write->end();
}

void MidiCopyEvent::prefixchannel(unsigned char channel)
{
  if (onlychannel >= 0 && onlychannel != channel)
    return;
  write->prefixchannel(channel);
}

void MidiCopyEvent::prefixport(unsigned char port)
{
  write->prefixport(port);
}

void MidiCopyEvent::text(int what, int len, char* whattext, unsigned char* txt)
{
  write->text(what, len, txt);
}


void MidiCopyEvent::error(const char* msg)
{
  printf("// error: %s\n", msg);
  exit(1);
}


void mididrum()
{
  int trackswritten = 0;

  MidiScan mainmidi(input);
  if (!mainmidi.getf())
  {
    perror(input);
    exit(1);
  }
  mainmidi.run();

  int drumcount = 0;
  for (int i = 0; i < 256;i++)
    if (drumused[i])
      drumcount++;

  if (drumcount == 0)
  {
    fprintf(stderr, "No drums found\n");
    if (!splitprograms)
      exit(1);
  }
  for (trackno = 0; trackno < trackcount; trackno++)
  {
    trackread[trackno] = new MidiCopyEvent(input, mainmidi.getf()); // sharing file!
    if (!trackread[trackno])
      break;
  }

  write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return;
  }

  write->head(1, 0, mainmidi.unitsperbeat_); // midi header format 1

  // first write all non drum tracks that are not splitted without changes
  for (trackno = 0; trackno < trackcount; trackno++)
  if (trackchannel[trackno] == NOCHANNEL ||
     (trackchannel[trackno] >= 0 && (!isdrum[trackchannel[trackno]] || !splitdrums)))
  {
    // write program splitted tracks later
    if (splitprograms && trackprograms[trackno] > 1 && !isdrum[trackchannel[trackno]])
      continue;

    write->track(); trackswritten++;
    trackread[trackno]->seek(trackpos[trackno]);
    while (trackread[trackno]->getcurpos() < trackend[trackno])
    {
      write->time(trackread[trackno]->getdelta());
      trackread[trackno]->runevent(trackend[trackno]);  // copy event
    }
    write->endtrack();
    // dont need this track anymore!
    delete trackread[trackno];
    trackread[trackno] = 0;
  }

  // and now the multi program tracks (option necessary)
  if (splitprograms)
  {
    if (mainmidi.version_ == VERSION_MULTICHANNEL && trackswritten == 0)
    {
      // create a tempo track
      write->track(); trackswritten++; // create a track for this program

      onlychannel = 16; // no channel depending events!
      onlynote = 128;   // no note event!
      onlyprogram = 128;  // no program event!
      assert(trackread[0] != 0);
      trackread[0]->seek(trackpos[0]-8);
      trackread[0]->runtrack(1);
      write->endtrack();
    }
    for (trackno = 0; trackno < trackcount; trackno++)
    if (trackchannel[trackno] < 0 || !isdrum[trackchannel[trackno]])
    if (trackprograms[trackno] > 1)
    {
      // split track into some programs

      // scan the programs
      mainmidi.seek(trackpos[trackno]-8);
      mainmidi.runtrack(trackno+1);

//      printf("track %d not splitted yet\n", trackno+1);
      for (int ch = 0; ch < 16; ch++)
      if (ch != drumchannel)
      for (int prog = 0; prog < 128; prog++)
      if (programused[ch][prog])
      {
	 write->track(); trackswritten++; // create a track for this program
	 mutenotes();

	 onlychannel = ch;
	 onlynote = -1;
	 onlyprogram = prog;
	 assert(trackread[trackno] != 0);
	 trackread[trackno]->seek(trackpos[trackno]-8);
	 trackread[trackno]->runtrack(trackno+1);
	 write->endtrack();
      }
    }
  }
  // and now the drum channels
  onlyprogram = -1;
  onlychannel = gm_drumchannel;

  // set to file pointers before first event
  firsttracksum = 0;
  for (trackno = 0; trackno < trackcount; trackno++)
  if (trackread[trackno])
  {
      // get time of first event
    trackread[trackno]->seek(trackpos[trackno]);
    firsttracktime[trackno] = trackread[trackno]->getdelta();
    firsttracksum += trackread[trackno]->getcurpos() - trackpos[trackno];
    firsttrackpos[trackno] = trackread[trackno]->getcurpos();
  }

  if (splitdrums)
  for (onlynote = 0; onlynote < MAXDRUMS; onlynote++)
  if (drumused[onlynote])
  {
    mutenotes();
    write->track();  trackswritten++;// create a track for this drum instrument
    for (int trk = 0; trk < trackcount; trk++)
    if (trackread[trk])
    {
      tracktime[trk] = firsttracktime[trk];
      trackpos[trk] = firsttrackpos[trk];
      tracksum = firsttracksum;
      for (int i = 0; i < 128; i++)
	trackctrl[trk][i] = -1;
    }
    for (int i = 0; i < 128; i++)  // control values of current track
      ctrl[i] = -1;
    unsigned long curtime = 0;
    int lastbesttrk = -1;
    for (;;)
    {
      int besttrk = -1;

      static int lastperc = -1;

      int perc;

      if (tracksum == 0)
	perc = 0;
      else
	perc = (int)((tracksum * 100) / tracktotal);

      if (perc != lastperc)
      {
	fprintf(stderr, "%d%%\r", perc);
	lastperc = perc;
      }
      // find track that has lowest time > 0
      for (trackno = 0; trackno < trackcount; trackno++)
      if (trackread[trackno] && tracktime[trackno] >= 0)
      {
	if (besttrk < 0 || tracktime[trackno] < tracktime[besttrk])
	{
	  besttrk = trackno;
	  if (tracktime[trackno] == curtime)
	    break;
	}
      }
      if (besttrk < 0)  // no more tracks with events
	break;
      trackno = besttrk;

      if (lastbesttrk != besttrk && lastbesttrk >= 0)
      {
	// see if controls have changed!
	for (int i = 0; i < 128; i++)
	if (ctrl[i] >= 0 && trackctrl[besttrk][i] >= 0 && ctrl[i] != trackctrl[besttrk][i])
	{
	  // other drum track set other controls
	  write->time(tracktime[besttrk] - curtime);
	  write->control(drumchannel, i, trackctrl[besttrk][i]);
	  ctrl[i] = trackctrl[besttrk][i];
	  curtime = tracktime[besttrk];
	}
      }
      tracksum -= trackpos[besttrk];
      write->time(tracktime[besttrk] - curtime);
      curtime = tracktime[besttrk];
      // copy event only if it is the drum note or not a note
      trackread[besttrk]->seek(trackpos[besttrk]);
      trackread[besttrk]->runevent(trackend[besttrk]);

      if (trackread[besttrk]->getcurpos() < trackend[besttrk])
      {
	tracktime[besttrk] += trackread[besttrk]->getdelta();
	trackpos[besttrk] = trackread[besttrk]->getcurpos();
	if (trackpos[besttrk] + 3 <= trackend[besttrk] &&
	    strcmp(trackread[besttrk]->get(3), "\xff\x2f") == 0) // meta event end track
	{
	  tracktime[besttrk] = -1;
	}
      }
      else
	  tracktime[besttrk] = -1;
      tracksum += trackpos[besttrk];
      lastbesttrk = besttrk;
    }

    write->endtrack(); // close the track
  }

  delete write;  // close midi file
  fprintf(stderr, "output written to %s\n", output);

  free(trackend);
  free(trackpos);
  free(tracktime);
  free(firsttrackpos);
  free(firsttracktime);
  free(trackread);
  free(trackchannel);
  free(trackctrl);
  free(trackprograms);
}

void usage()
{
  printf("usage: mididrum [-drum][-program][-both] srcfile.mid dstfile.mid\n");
  printf("splits drum instruments into separate tracks\n");
  printf("This program only accepts format 0 and 1 midi files.\n");
  printf("-version\tget program version\n");
  printf("-drum\t\tsplit drum tracks by drum instruments\n");
  printf("-program\tsplit non-drum tracks by programs\n");
  printf("-both\t\tsplit drums and programs\n");
  printf("Format 0 midi files (1 track) can only be splitted with option -both\n");
  exit(1);
}

int main(int argc, char** argv)
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (strncmp(*argv, "-program", 2) == 0)
    {
      splitprograms = 1;
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-drum", 2) == 0)
    {
      splitdrums = 1;
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-both", 2) == 0)
    {
      splitprograms = splitdrums = 1;
      argc--; argv++; continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert midi to same file\n");
    return 1;
  }

  if (!splitprograms && !splitdrums)
    splitdrums = 1;

  mididrum();
  return 0;
}
