#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>

static char* version = "midinote v2.0 by Gnter Nagler (" __DATE__ ")";

char* inputname = 0;
char* outputname = 0;
FILE* outputf = 0;

int values = 0; // cryptical values or readable information
     // values are easier for import in programs
int header = 1;
     // show what columns mean
     // import in programs often don't need this information

class EventInfo
{
public:
  EventInfo();

  virtual const int EventType() { return 0; }
  virtual const char* EventName() { return NULL; }
  virtual int ready() { return 1; }

  int track_;
  unsigned char channel_;
  unsigned long startunit_;
  unsigned long startms_;

  virtual void print();
};

class NoteInfo : public EventInfo
{
public:
  NoteInfo();

  virtual const int EventType() { return 'N'; }
  virtual const char* EventName() { return "Note"; }
  virtual int ready() { return ready_; }

  unsigned long endunit_;
  unsigned long endms_;
  unsigned char note_;
  unsigned char vel_;
  char ready_;

  virtual void print();
};

class HeaderInfo : public EventInfo
{
public:
  HeaderInfo(int unitsperbeat, int version, int trackcount);
  HeaderInfo(MidiRead*);

  virtual const int EventType() { return 'H'; }
  virtual const char* EventName() { return "Header"; }

  virtual void print();

  int unitsperbeat_, version_, trackcount_;
};

class TempoInfo : public EventInfo
{
public:
  TempoInfo(float beatspermin);

  virtual const int EventType() { return 'T'; }
  virtual const char* EventName() { return "Tempo"; }

  virtual void print();

  float beatspermin_;
};

class TactInfo : public EventInfo
{
public:
  TactInfo(int nomin, int denom);

  virtual const int EventType() { return 'B'; }
  virtual const char* EventName() { return "Bar"; }

  virtual void print();

  int nomin_, denom_; // e.g. 3/4
};

class MidiNotes : public MidiSerial
{
public:
  MidiNotes(const char* filename);
  ~MidiNotes();

  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);

  virtual void tempo(unsigned long microsecperbeat);
  virtual void tact(int nom, int denom, int unitsperbeat, int notes32perbeat);

  void addevent(EventInfo* event);

  void startevent(EventInfo* event);
  void stopnote(NoteInfo* note);

  virtual void printevent(EventInfo* event);

  virtual void endmidi();
  void print();

protected:
  unsigned long printunit_;
  long recordcnt_;

  EventInfo** list_;
  int maxcount_, count_;

  int curtact_nomin_, curtact_denom_, curtact_units_;
  long lasttact_unit_;

  int curtactunits();

  int findnote(NoteInfo* note);
  void removeevent(int i);
};

void outofmemory()
{
  fprintf(stderr, "out of memory\n");
  exit(1);
}

EventInfo::EventInfo()
{
  channel_ = 0;
  track_ = 0;
  startunit_ = 0;
  startms_ = 0;
}

void EventInfo::print()
{
  if (values)
    fprintf(outputf, "%c\t%ld\t%ld\t%d\t%d", EventType(), startunit_, startms_, track_, channel_);
  else
    fprintf(outputf, "%s\t%ld\t%ldms\tT%d\tC%d", EventName(), startunit_, startms_, track_, channel_);
}

HeaderInfo::HeaderInfo(int unitsperbeat, int version, int trackcount)
{
  unitsperbeat_ = unitsperbeat;
  version_ = version;
  trackcount_ = trackcount;
}

HeaderInfo::HeaderInfo(MidiRead* midi)
{
  unitsperbeat_ = midi->unitsperbeat_;
  version_ = midi->version_;
  trackcount_ = midi->tracks_;
}

void HeaderInfo::print()
{
  EventInfo::print();
  if (values)
    fprintf(outputf, "\t%d\t%d\t%d\n", version_, unitsperbeat_, trackcount_);
  else
    fprintf(outputf, "\tVer%d\t%dunits\t%dtracks\n", version_, unitsperbeat_, trackcount_);
}

TempoInfo::TempoInfo(float beatspermin)
{
  beatspermin_ = beatspermin;
}

void TempoInfo::print()
{
  EventInfo::print();
  if (values)
    fprintf(outputf, "\t%2.2f\n", beatspermin_);
  else
    fprintf(outputf, "\t%2.2fbpm\n", beatspermin_);
}

TactInfo::TactInfo(int nomin, int denom)
{
  nomin_ = nomin;
  denom_ = denom;
}

void TactInfo::print()
{
  EventInfo::print();
  if (values)
    fprintf(outputf, "\t%d\t%d\n", nomin_, denom_);
  else
    fprintf(outputf, "\t%d/%d\n", nomin_, denom_);
}


NoteInfo::NoteInfo()
{
  endunit_ = 0;
  endms_ = 0;
  note_ = 0;
  vel_ = 0;
  ready_ = 0;
}

void NoteInfo::print()
{
  EventInfo::print();
  if (values)
  {
    fprintf(outputf, "\t%ld\t%ld\t%d\t%d\n",
      endunit_ - startunit_,
      endms_ - startms_,
      note_,
      vel_);
  }
  else
  {
    fprintf(outputf, "\t%ldunits\t%ldms\t%s\tVel%d\n",
      endunit_ - startunit_,
      endms_ - startms_,
      MidiRead::notename(note_),
      vel_);
  }
}

MidiNotes::MidiNotes(const char* filename) : MidiSerial(filename)
{
  list_ = 0;
  maxcount_ = count_ = 0;
  printunit_ = 0;
  recordcnt_ = 0;
  curtact_nomin_ = curtact_denom_ = 4;
  curtact_units_ = curtactunits();
  lasttact_unit_ = -1;
}

int MidiNotes::curtactunits()
{
  return curtact_nomin_ * unitsperbeat_ * 4 / curtact_denom_;
}

MidiNotes::~MidiNotes()
{
  for (int i = 0; i < count_; i++)
    delete list_[i];
  maxcount_ = count_ = 0;
  if (list_)
     free(list_);
}

void MidiNotes::endmidi()
{
  print();
  if (count_ > 0)
    fprintf(stderr, "Warning: %d notes were not stopped\n", count_);
  for (int i = 0; i < count_; i++)
  {
    if (list_[i]->EventType() != 'N')
      continue;
    NoteInfo* note = (NoteInfo*)list_[i];
    note->endunit_ = getcurunit();
    note->endms_ = getcurmillisec();
    note->ready_ = 1;
  }
  print();
}

void MidiNotes::startevent(EventInfo* note)
{
  if (!note)
    return;
  addevent(note);
}

void MidiNotes::addevent(EventInfo* note)
{
  if (!note)
    return;
  if (count_ == maxcount_)
  {
    if (maxcount_ >= 16000)
      list_ = 0;
    else
    {
      maxcount_ += 100;
      list_ = (EventInfo**)realloc(list_, maxcount_ * sizeof(EventInfo*));
    }
    if (!list_)
      outofmemory();
  }
  list_[count_++] = note;
}

int MidiNotes::findnote(NoteInfo* note)
{
int besti = -1;

  for (int i = 0; i < count_; i++)
  if (list_[i]->EventType() == 'N')
  {
    NoteInfo* n = (NoteInfo*)list_[i];
    if (!n->ready() && n->note_ == note->note_ &&
      n->channel_ == note->channel_)
    {
      if (n->track_ == note->track_)
	return i;
      besti = i; // track not matching
    }
  }
  return besti;
}

void MidiNotes::removeevent(int i)
{
  if (i  < 0 || i >= count_)
    return;
  delete list_[i]; list_[i] = NULL;
  if (i != count_-1)
    list_[i] = list_[count_-1];
  list_[count_-1] = NULL;
  count_ --;

  if (maxcount_ - 100 > count_)
  {
    maxcount_ = count_+100;
    list_ = (EventInfo**)realloc(list_, maxcount_ * sizeof(EventInfo*));
    if (!list_)
      outofmemory();
  }
}

void MidiNotes::stopnote(NoteInfo* note)
{
  int i = findnote(note);
  if (i < 0)
    return;

  NoteInfo* n = (NoteInfo*)list_[i];

  n->endunit_ = note->endunit_;
  n->endms_ = note->endms_;
  n->ready_ = 1;

  print();
}

void MidiNotes::print()
{
unsigned long nextunit;
int besti = -1;
int printcnt;

  do
  {
    printcnt = 0;
    besti = -1;
    for (int i = 0; i < count_; i++)
    {
      if (list_[i]->ready() && list_[i]->startunit_ == printunit_)
      {
	printevent(list_[i]); printcnt++;
	removeevent(i);
        i--; // process current note again
        continue;
      }
      if (besti < 0 || (i <count_ && list_[i]->startunit_ < nextunit))
      {
	nextunit = list_[i]->startunit_;
	besti = i;
      }
    }
    if (besti < 0)  // no next unit found (maybe no notes available)
      break;
    if (printunit_ == nextunit)
      break;
    printunit_ = nextunit;
  } while (1);
}

void MidiNotes::printevent(EventInfo* event)
{
  if (!event)
     return;

  if (event->startunit_ < printunit_)
     fprintf(stderr, "warning: midinotes are not sorted by start time!\n");

  if (recordcnt_ == 0 && header)
  {
    fprintf(outputf, "Unit\tAt\tTrack\tChannel\tCommand\tParams\n");
  }
  event->print();
  recordcnt_++;
}

void MidiNotes::noteon(int channel, int note, int vel)
{
  NoteInfo* ninfo = new NoteInfo;
  if (!ninfo)
  {
    outofmemory();
    return;
  }

  // see if note is currently playing
  ninfo->track_ = trackno_;
  ninfo->endunit_ = getcurunit();
  ninfo->endms_ = getcurmillisec();
  ninfo->channel_ = channel+1;
  ninfo->note_ = note;
  ninfo->vel_ = vel;
  if (findnote(ninfo) >= 0)
    stopnote(ninfo);

  ninfo->track_ = trackno_;
  ninfo->startunit_ = getcurunit();
  ninfo->startms_ = getcurmillisec();
  ninfo->channel_ = channel+1;
  ninfo->note_ = note;
  ninfo->vel_ = vel;

//  printf("%ld ON[%d] %d\n", ninfo->startunit_, ninfo->channel_, ninfo->note_);
  startevent(ninfo);
}

void MidiNotes::noteoff(int channel, int note, int vel)
{
  NoteInfo* ninfo = new NoteInfo;
  if (!ninfo)
  {
    outofmemory();
    return;
  }
  ninfo->track_ = trackno_;
  ninfo->endunit_ = getcurunit();
  ninfo->endms_ = getcurmillisec();
  ninfo->channel_ = channel+1;
  ninfo->note_ = note;
  ninfo->vel_ = vel;

//  printf("%ld OFF[%d] %d\n", ninfo->endunit_, ninfo->channel_, ninfo->note_);
  stopnote(ninfo);

  delete ninfo;
}

void MidiNotes::tempo(unsigned long ticks)
{
  TempoInfo* t = new TempoInfo(bpm(ticks));
  if (!t)
    outofmemory();

  t->startunit_ = getcurunit();
  t->startms_ = getcurmillisec();
  t->track_ = trackno_;
  startevent(t);
}

void MidiNotes::tact(int nom, int denom, int , int )
{
  curtact_nomin_ = nom;
  curtact_denom_ = denom;
  curtact_units_ = curtactunits();
  TactInfo* tact = new TactInfo(nom, denom);
  if (!tact)
      outofmemory();
  tact->startunit_ = getcurunit();
  tact->startms_ = getcurmillisec();
  tact->track_ = trackno_;
  startevent(tact);
}

void usage()
{
  printf("usage: midinote [-noheader] [-values] input.mid [output.txt]\n");
  printf("-noheader\tdon't print header that explains the columns\n");
  printf("-values\t\tprint cryptical values for easier import into programs\n");
  printf("input.mid\tget notes sorted by start time from this midi file\n");
  printf("output.txt\twrite output into a text file (default: standard output)\n");
}

int main(int argc, char** argv)
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strcmp(*argv, "-version") == 0)
    {
      fprintf(stderr, "%s\n", version);
    }
    if (strncmp(*argv, "-noheader", 2) == 0)
    {
      header = 0;
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-values", 2) == 0)
    {
      values = 1;
      argc--; argv++; continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
  }
  if (argc == 0)
  {
    usage();
    return 1;
  }
  inputname = *argv++; argc--;
  if (argc > 0)
  {
    outputname = *argv++; argc--;
  }

  MidiNotes* notes = new MidiNotes(inputname);
  if (!notes)
    outofmemory();

  if (outputname)
  {
    outputf = fopen(outputname, "w");
    if (!outputf)
    {
      perror(outputname);
      return 1;
    }
  }
  else
    outputf = stdout;

  notes->startevent(new HeaderInfo(notes));
  notes->run();
  delete notes;
  if (outputname)
    fclose(outputf);
  return 0;
}