// midipat v1.0 written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef __MSDOS__
#define WRITE_BINARY  "wb"
#define READ_BINARY   "rb"
#else
#define WRITE_BINARY  "w"
#define READ_BINARY   "r"
#endif

char* input = 0;
char* output = 0;
char* patchname = "midipat.pat";

int quiet = 0;
int isdrum[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0 };
int patchannel[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

#define MAXPROGRAMS  128
#define MAXDRUMS     128
#define MAXNOTES     128
#define DRUMS  MAXPROGRAMS

class PATCH
{
public:
  PATCH();

  int hbank, lbank;
  int program;
};

PATCH::PATCH()
{
  hbank = lbank = program = -1;
}

PATCH patchprog[128+1];  // 128 programs and 1 drum
PATCH patchdrum[128];


class MidiPatch : public MidiCopy
{
public:
  MidiPatch(char* name);

  virtual void track(int trackno, long length, int channel);

  virtual void program(int prg, int channel);
  virtual void highbank(int channel, int val);
  virtual void lowbank(int channel, int val);
  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void error(const char* msg);

  virtual void percent(int perc);

  void patchprogram(int channel, int prg);
  void patchnote(int channel, int &note);
};

MidiPatch::MidiPatch(char* name) : MidiCopy(name)
{
  options_ = OPTION_NOSYSEVENTS;
}

void MidiPatch::patchnote(int channel, int &note)
{
  if (!patchannel[channel])
    return;
  if (isdrum[channel])
  {
  PATCH &p = patchdrum[note];

    if (p.program >= 0)
      note = p.program;
  }
}

void MidiPatch::patchprogram(int channel, int prg)
{
  if (!patchannel[channel])
  {
    MidiCopy::program(channel, prg);
    return;
  }

  PATCH &p = isdrum[channel] ? patchprog[DRUMS] : patchprog[prg];
  if (p.hbank >= 0)
    MidiCopy::highbank(channel, p.hbank);
  if (p.lbank >= 0)
    MidiCopy::lowbank(channel, p.lbank);
  MidiCopy::program(channel, (p.program >= 0) ? p.program : prg);
}

void MidiPatch::track(int trackno, long length, int channel)
{
  MidiCopy::track(trackno, length, channel);
  if (channel >= 0 && isdrum[channel])  // initialize drum channel
    patchprogram(channel, DRUMS);
  else if (channel == MULTICHANNEL)
  {
    for (int c = 0; c < 16; c++)
    if (isdrum[c])
      patchprogram(c, DRUMS);
  }
}

void MidiPatch::program(int channel, int prg)
{
  patchprogram(channel, prg);
}

void MidiPatch::highbank(int /*channel*/, int /*val*/)
{
  // ignore highbank
}

void MidiPatch::lowbank(int /*channel*/, int /*val*/)
{
  // ignore lowbank
}

void MidiPatch::noteon(int channel, int note, int vel)
{
  patchnote(channel, note);
  MidiCopy::noteon(channel, note, vel);
}

void MidiPatch::noteoff(int channel, int note, int vel)
{
  patchnote(channel, note);
  MidiCopy::noteoff(channel, note, vel);
}

void MidiPatch::percent(int perc)
{
  if (!quiet)
    fprintf(stderr, "%-3d%%\r", perc);
}

void MidiPatch::error(const char* msg)
{
  printf("// error: %s\n", msg);
}

void usage()
{
  fprintf(stderr, "midipat translates GM programs into device dependent programs\n");
  fprintf(stderr, "usage: midipat [-P] [-D] [-patch file.pat] [-c #][-drum #] [-quiet] file.mid patched.mid\n");
  fprintf(stderr, "-P\t\tshow list of GM program names\n");
  fprintf(stderr, "-D\t\tshow list of GM drum names\n");
  fprintf(stderr, "-patch file.pat\tload patch from file.pat (default is midipat.pat)\n");
  fprintf(stderr, "-c #\t\tmodify programs on channel (1-16, default: all channels)\n");
  fprintf(stderr, "-drum #\t\tdeclare channel(s) # (1-16) as drum channel (default is 10)\n");
  fprintf(stderr, "-quiet\t\tbe quiet! does not write percentage info\n");
  exit(1);
}

void copyfile(char* input, char* output)
{
FILE* i = fopen(input, READ_BINARY);

  if (!i)
  {
    perror(input);
    return;
  }
  FILE* o = fopen(output, WRITE_BINARY);
  if (!o)
  {
    fclose(i);
    perror(output);
    return;
  }
  static char buf[4096];
  int n;
  long pos = 0;

  while (fread(buf, sizeof(buf), 1, i) == 1)
  {
    fwrite(buf, sizeof(buf), 1, o);
    pos += sizeof(buf);
  }
  fseek(i, pos, SEEK_SET);
  fseek(o, pos, SEEK_SET);
  while ((n = fread(buf, 1, sizeof(buf), i)) > 0)
    fwrite(buf, n, 1, o);

  fclose(i);
  fclose(o);
}

char* loadline(FILE* f, char* s)
{
int c;
char* t = s;

  while ((c = fgetc(f)) != EOF && c != '\n')
  {
    if (c == '\r')
      continue;
    *t++ = c;
  }
  *t = 0;
  if (c == EOF && s == t)
    return 0;
  return s;
}

char* getword(char* &s)
{
static char word[256];
char* w = word;

  while (*s != 0 && isspace(*s))
    s++;
  w = word;
  while (*s != 0 && !isspace(*s))
    *w++ = *s++;
  *w = '\0';
  return word;
}

static char* GMProg[128] =
{
  "Piano", "BritePiano", "HammerPiano", "HonkeyTonk", "NewTines", "DigiPiano", "Harpsicord", "Clav",
  "Celesta", "Glocken", "MusicBox", "Vibes", "Marimba", "Xylophon", "Tubular", "Santur",
  "FullOrgan", "PercOrgan", "BX-3Organ", "ChurchPipe", "Positive", "Musette", "Harmonica", "Tango",
  "ClassicGtr", "A.Guitar", "JazzGuitar", "CleanGtr", "MuteGuitar", "OverDrive", "DistGuitar", "RockMonics",
  "JazzBass", "DeepBass", "PickBass", "FretLess", "SlapBass1", "SlapBass2", "SynthBass1", "SynthBass2",
  "Violin", "Viola", "Cello", "ContraBass", "TremoloStr", "Pizzicato", "Harp", "Timpani",
  "Marcato", "SlowString", "AnalogPad", "StringPad", "Choir", "DooVoice", "Voices", "OrchHit",
  "Trumpet", "Trombone", "Tuba", "MutedTrumpet", "FrenchHorn", "Brass", "SynBrass1", "SynBrass2",
  "SopranoSax", "AltoSax", "TenorSax", "BariSax", "SweetOboe", "EnglishHorn", "BasoonOboe", "Clarinet",
  "Piccolo", "Flute", "Recorder", "PanFlute", "Bottle", "Shakuhachi","Whistle", "Ocarina",
  "SquareWave", "SawWave", "SynCalinope", "SynChiff", "Charang", "AirChorus", "Rezzo4ths", "Bass&Lead",
  "Fantasia", "WarmPad", "PolyPad", "GhostPad", "BowedGlas", "MetalPad", "HaloPad", "Sweep",
  "IceRain", "SoundTrack", "Crystal", "Atmosphere", "Brightness", "Goblin", "EchoDrop", "StarTheme",
  "Sitar", "Banjo", "Shamisen", "Koto", "Kalimba","Scotland","Fiddle", "Shanai",
  "MetalBell", "Agogo", "SteelDrums", "Woodblock", "Taiko", "Tom", "SynthTom", "RevCymbal",
  "FretNoise", "NoiseChiff", "Seashore", "Birds", "Telephone", "Helicopter", "Stadium!!", "GunShot"
};

static char* GMDrumProg[MAXDRUMS] =
{
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 
  0, 0, 0, 0,"Snare3","RockKick","OpenHH","FatKick",
  "Timbales","Snare1","RollSnare1","RealKick","ProcesKick","SideStick","RockSnare","HandClaps",
  "LightSnare","TomLo","TiteHH","TomLo","PedalHH","TomLo","OpenHH","TomHi",
  "TomHi","CrashCym","TomHi","RideEdge","ChinaCym","RideCup","Tambourine","SplashCym",
  "Cowbell","CrashCym","Vibraslap","RideCym1","HiBongo","LoBongo","MuteConga","OpenConga",
  "OpenConga","HiTimbal","LoTimbal","Agogo","Agogo","Cabasa","Maracas","WhistleS",
  "WhistleL","GuiroS","GuiroL","Claves","WoodBlock2","WoodBlock3","MuteCuica","OpenCuica",
  "MuteTriang","OpenTriang","Cabasa","JingleBell","BellTree","Castanet","SideStick","TaikoLo",
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0,
};

static char* GMNote[MAXNOTES] =
{
  0,0,0,0,0,0,0,0,0,0,0,0,
  "C0","C#0","D0","D#0","E0","F0","F#0","G0","G#0","A0","A#0","B0",
  "C1","C#1","D1","D#1","E1","F1","F#1","G1","G#1","A1","A#1","B1",
  "C2","C#2","D2","D#2","E2","F2","F#2","G2","G#2","A2","A#2","B2",
  "C3","C#3","D3","D#3","E3","F3","F#3","G3","G#3","A3","A#3","B3",
  "C4","C#4","D4","D#4","E4","F4","F#4","G4","G#4","A4","A#4","B4",
  "C5","C#5","D5","D#5","E5","F5","F#5","G5","G#5","A5","A#5","B5",
  "C6","C#6","D6","D#6","E6","F6","F#6","G6","G#6","A6","A#6","B6",
  "C7","C#7","D7","D#7","E7","F7","F#7","G7","G#7","A7","A#7","B7",
  "C8","C#8","D8","D#8","E8","F8","F#8","G8","G#8","A8","A#8","B8",
  "C9","C#9","D9","D#9","E9","F9","F#9","G9"
};

int equali(char* s1, char* s2)
{
  if (!s1 || !s2)
    return 0;
  while (*s1 != 0 && *s2 != 0)
    if (toupper(*s1++) != toupper(*s2++))
      return 0;
  if (*s1 != *s2)
    return 0;
  return 1;
}

int scanprogram(char* name, int &program, int &drum)
{
  program = drum = -1;

  if (strlen(name) >= 2 &&
      ((toupper(name[0]) == 'P' && isdigit(name[1])) ||
      (toupper(name[0]) == 'D' && toupper(name[1]) == 'R' && isdigit(name[2])))
     )
  {
    int n = 0;
    char* s = name+1;

    if (toupper(name[0]) != 'P')
      s++;
    for (; *s && isdigit(*s); s++)
      n = n * 10 + (*s - '0');
    if (*s == '\0')
    {
      if (toupper(*name) == 'P' && n < MAXPROGRAMS)
      {
	drum = 0;
	program = n;
	return 1;
      }
      else if (toupper(*name) == 'D' && toupper(name[1]) == 'R' && n < MAXDRUMS)
      {
	drum = 1;
	program = n;
	return 1;
      }
    }
  }
  if (equali(name, "Drums"))
  {
    program = DRUMS; // all drums
    drum = 0;
    return 1;
  }
  for (int i = 0; i < MAXPROGRAMS; i++)
    if (equali(name, GMProg[i]))
    {
      program = i;
      drum = 0;
      return 1;
    }
  for (i = 0; i < MAXDRUMS; i++)
    if (equali(name, GMDrumProg[i]) || equali(name, GMNote[i]))
    {
      program = i;
      drum = 1;
      return 1;
    }
  return 0;
}

void shownames(int drums)
{
int n = drums ? MAXDRUMS : MAXPROGRAMS;
char** list = drums ? GMDrumProg : GMProg;
char *c = drums ? "Dr" : "P";

  printf("%ss:\n", drums ? "Drum" : "Program");
  for (int i = 0; i < n; i++)
  if (list[i])
    printf("%s%03d=%s\n", c, i, list[i]);
  printf("You can use numbers instead of names too: %s0-%s%d\n", c, c, n-1);
}

int getnumber(char* word)
{
char* s = word;
int v = 0;

  if (*s == '-')
    return -1;
  if (isdigit(*s))
  {
    while (*s && isdigit(*s))
      v = v * 10 + (*s++ - '0');
  }
  else if (*s == '$' || strncmp(s, "0x",2) == 0 ||strncmp(s, "0X", 2) == 0)
  {
    if (*s == '$')
      s++;
    else
      s+=2;
    while (*s && isxdigit(*s))
    {
      if (isdigit(*s))
	v = (v << 4) + (*s - '0');
      else
	v = (v << 4) + (toupper(*s) - 'A' + 10);
      s++;
    }
  }
  else
  {
    fprintf(stderr, "Number expected instead of %s\n", word);
    return -1;
  }
  if (*s != '\0')
    fprintf(stderr, "Invalid characters in number %s\n", word);
  return v;
}

int loadpatchtable(char* filename)
{
FILE* f = fopen(filename, READ_BINARY);
static char line[256];
char *lp;
int program, drum;
int invalidprogram = 0;

  if (!f)
  {
    perror(filename);
    return 0;
  }

  while (loadline(f, line))
  {
  char* word;

    lp = line;

    word = strchr(lp, ';');  // remove comment!
    if (word)
      *word = '\0';

    // get gm program or gm drum name
    word = getword(lp);
    if (*word == 0)     // empty line!
      continue;
    if (!scanprogram(word, program, drum))
    {
      fprintf(stderr, "Invalid program used in patch file: %s\n", word);
      invalidprogram = 1;
      continue;
    }

    if (drum)
    {
      int note = -1;

      // get note value (can be empty (-))
      word = getword(lp);
      if (!*word)
      {
	fprintf(stderr, "invalid line format: %s\n", line);
	continue;
      }
      if (strcmp(word, "-") == 0)
	note = -1;
      else
      {
      int d;

	if (!scanprogram(word, note, d))
	  note = getnumber(word);
	if (note >= MAXDRUMS)
	{
	  fprintf(stderr, "%s: note value should be in range of 0-%d\n", word, MAXDRUMS-1);
	  note = -1;
	}
      }
      patchdrum[program].hbank = -1;
      patchdrum[program].lbank = -1;
      patchdrum[program].program = note;
    }
    else
    {
      int hbank = -1;
      // get high bank value (can be empty (-))
      word = getword(lp);
      if (!*word)
      {
	fprintf(stderr, "invalid line format: %s\n", line);
	continue;
      }
      if (strcmp(word, "-") == 0)
	hbank = -1;
      else
      {
	hbank = getnumber(word);
	if (hbank >= 128)
	{
	  fprintf(stderr, "%s: high bank should be in range of 0-127\n", word);
	  hbank = -1;
	}
      }

      // get low bank value (can be empty (-))
      int lbank = -1;
      // get high bank value (can be empty (-))
      word = getword(lp);
      if (!*word)
      {
	fprintf(stderr, "invalid line format: %s\n", line);
	continue;
      }
      if (strcmp(word, "-") == 0)
	lbank = -1;
      else
      {
	lbank = getnumber(word);
	if (lbank >= 128)
	{
	  fprintf(stderr, "%s: low bank should be in range of 0-127\n", word);
	  lbank = -1;
	}
      }

      // get program number
      int nprog = -1;
      // get high bank value (can be empty (-))
      word = getword(lp);
      if (!*word)
      {
	fprintf(stderr, "invalid line format: %s\n", line);
	continue;
      }
      if (strcmp(word, "-") == 0)
	nprog = -1;
      else
      {
      int d;

	if (!scanprogram(word, nprog, d))
	  nprog = getnumber(word);

	if (nprog >= 128)
	{
	  fprintf(stderr, "%s: program should be in range of 0-127\n", word);
	  nprog = -1;
	}
      }
      patchprog[program].hbank = hbank;
      patchprog[program].lbank = lbank;
      patchprog[program].program = nprog;
    }
  }
  fclose(f);
  if (invalidprogram)
  {
    fprintf(stderr, "Use option -P or -D to get a list of names\n");
    return 0;
  }
  return 1;
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-c", 2) == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
      {
	fprintf(stderr, "option -c needs channel parameter (1-16)\n");
	return 1;
      }
      else
      {
	int channel = atoi(*argv);
	if (channel <= 0 || channel > 16)
	{
	  fprintf(stderr, "channel parameter out of range (1-16) %s\n", *argv);
	  return 1;
	}
	patchannel[channel-1] = 1;
      }
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-p", 2) == 0)
    {
      argc--; argv++;
      if (argc == 0)
      {
	fprintf(stderr, "option -patch needs filename parameter\n");
	continue;
      }
      patchname = *argv++; argc--;
      continue;
    }
    if (strncmp(*argv, "-quiet", 2) == 0)
    {
      quiet = 1;
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-P") == 0)
    {
      shownames(0);
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-D") == 0)
    {
      shownames(1);
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-drum", 2) == 0)
    {
      argc--; argv++;
      if (argc == 0 || !isdigit(**argv))
      {
	fprintf(stderr, "option -drum needs channel number(s) as parameter\n");
	continue;
      }
      while (argc > 0 && isdigit(**argv))
      {
	int ch = atoi(*argv); 
	if (ch <= 0 || ch > 16)
	  fprintf(stderr, "invalid channel %s, must be 1-16\n", *argv);
	else
	  isdrum[ch-1] = 1;
	argc--; argv++;
      }
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  for (int i = 0; i < 16; i++)
  if (patchannel[i])
    break;
  if (i >= 16) // no channel set => modify all channels
  {
    for (i = 0; i < 16; i++)
      patchannel[i] = 1;
  }

  if (!loadpatchtable(patchname))
    return 0;

  if (argc < 2)
    usage();

  input = argv[0];
  output = argv[1];
  if (strcmp(input, output) == 0)
  {
    fprintf(stderr, "cannot convert midi to same file\n");
    return 1;
  }

  MidiPatch midi(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }
  MidiWrite* write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return 1;
  }
  if (!write->getf())
  {
    perror(output);
    return 1;
  }

  midi.setoutput(write);

  midi.options_ = OPTION_NOSYSEVENTS;

  argc--; argv++;
  if (!midi.run())
    fprintf(stderr, "%s: midi read error at %04lX\n", input, midi.getpos());

  delete write;

  return 0;
}


