// midiprog written by Gnter Nagler 1997 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "progname.h"

static char* version = "midiprog v1.0 by Gnter Nagler (" __DATE__ ")";

char* input = 0;

enum midimodes { GM, XG, GS } mode = GM;

class MidiProgram : public MidiSerial
{
public:
  MidiProgram(char* name);

  virtual void program(int prg, int channel);
  virtual void highbank(int channel, int val);
  virtual void lowbank(int channel, int val);
  virtual void noteon(int channel, int note, int vel);

  virtual void xgreset();
  virtual void gmreset();
  virtual void gsreset();

  virtual void error(const char* msg);
  virtual void warning(const char* msg);

  char prog[16];
  char hbank[16];
  char lbank[16];
};

int curhbank[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
int curlbank[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
int curprog[16] = { -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };

void useprogram(int channel, int hbank, int lbank, int prog)
{
  if (curprog[channel] == prog && curhbank[channel] == hbank && curlbank[channel] == lbank)
    return;
  curprog[channel] = prog;
  curhbank[channel] = hbank;
  curlbank[channel] = lbank;

  const char* name = 0;
  const char* modename = 0;
  switch(mode)
  {
  case GM: default:
	name = gmpatch(channel, hbank, lbank, prog);
	modename = "GM";
	break;
  case GS:
	name = gspatch(channel, hbank, lbank, prog);
	modename = "GS";
	break;
  case XG:
	name = xgpatch(channel, hbank, lbank, prog);
	modename = "XG";
	break;
  }

  printf("mode=%s channel=%-2d program=%02X hbank=%02X lbank=%02X name=%s\n",
    modename, SAYCHANNEL(channel), prog, hbank, lbank, name ? name : "unknown");
}

MidiProgram::MidiProgram(char* name) : MidiSerial(name)
{
  for (int c = 0; c < 16; c++)
  {
    prog[c] = 0;
    hbank[c] = 0;
    lbank[c] = 0;
  }
}

void MidiProgram::program(int channel, int prg)
{
  prog[channel] = prg;
}

void MidiProgram::highbank(int channel, int val)
{
  hbank[channel] = val;
}

void MidiProgram::lowbank(int channel, int val)
{
  lbank[channel] = val;
}

void MidiProgram::noteon(int channel, int note, int vel)
{
  useprogram(channel, hbank[channel], lbank[channel], prog[channel]);
}

void MidiProgram::gmreset()
{
  mode = GM;
}

void MidiProgram::gsreset()
{
  mode = GS;
}

void MidiProgram::xgreset()
{
  mode = XG;
}


void MidiProgram::error(const char* msg)
{
  printf("Error: %s\n", msg);
  exit(1);
}

void MidiProgram::warning(const char* msg)
{
  printf("Warning: %s\n", msg);
}

void usage()
{
  fprintf(stderr, "MidiProgram searches all programs that are used\n");
  fprintf(stderr, "usage: MidiProgram [-xg][-gs][-gm] file.mid\n");
  fprintf(stderr, "-xg\tassume XG mode\n");
  fprintf(stderr, "-gs\tassume GS mode\n");
  fprintf(stderr, "-gm\tassume GM mode\n");
  exit(1);
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (strcmp(*argv, "-xg") == 0)
    {
      mode = XG;
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-gs") == 0)
    {
      mode = GS;
      argc--; argv++; continue;
    }
    if (strcmp(*argv, "-gm") == 0)
    {
      mode = GM;
      argc--; argv++; continue;
    }

    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 1)
    usage();

  input = argv[0];
  MidiProgram midi(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }
  argc--; argv++;

  midi.options_ = OPTION_NOMETAEVENTS;
  if (!midi.run())
    fprintf(stderr, "%s: midi read error at %04lX\n", input, midi.getpos());

  return 0;
}
