// midivol written by Gnter Nagler 1995 (gnagler@ihm.tu-graz.ac.at)
#include "midiio.hpp"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static char* version = "midivol v1.0 by Gnter Nagler (" __DATE__ ")";

char* input = 0;
char* output = 0;

#define DEFAULTLEVEL 13000

int expectedlevel = DEFAULTLEVEL;
int maxlevel = 0;

int vollevel(int vol, int vel)
{
  return vol * vel;
}

int scalevel(int vol, int vel, int newvol)
{
int newlevel;

  if (newvol <= 0)
    return vel;
  newlevel = (int) ((long(vol) * vel * expectedlevel) / maxlevel);
  int newvel = newlevel / newvol;
  if (newvel > 127)
    return 127;
  if (newvel < 0)
    return 0;
  return newvel;
}

class MidiScanVolume : public MidiRead
{
public:
  MidiScanVolume(char* name);

  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void volume(int channel, int vol);

  virtual void endmidi();

  virtual void error(const char* msg);
  virtual void warning(const char* msg);

  virtual void percent(int perc);

  int maxvolume_;
  int maxlevel_;
  int minlevel_;
  int avglevel_;

protected:
  long n_;
  long sumlevel_;
  int volume_;
};

MidiScanVolume::MidiScanVolume(char* name) : MidiRead(name)
{
  volume_ = 90; // default GM volume
  minlevel_ = 128*128;
  maxlevel_ = 0;
  n_ = 0;
  sumlevel_ = 0;
  maxvolume_ = 0;
}

void MidiScanVolume::volume(int channel, int vol)
{
  volume_ = vol;
}

void MidiScanVolume::noteon(int channel, int note, int vel)
{
  int level = vollevel(volume_, vel);
  if (level < minlevel_)
    minlevel_ = level;
  else if (level > maxlevel_)
    maxlevel_ = level;

  sumlevel_+=level;
  n_++;

  if (volume_ > maxvolume_)
    maxvolume_ = volume_;

  MidiRead::noteon(channel, note, vel);
}

void MidiScanVolume::noteoff(int channel, int note, int vel)
{
  int level = vollevel(volume_, vel);
  if (level < minlevel_)
    minlevel_ = level;
  else if (level > maxlevel_)
    maxlevel_ = level;

  MidiRead::noteoff(channel, note, vel);
}

void MidiScanVolume::endmidi()
{
  if (n_ > 0)
  {
    avglevel_ = (int)(sumlevel_ / n_);
  }
  else
    avglevel_ = 0;
}

void MidiScanVolume::percent(int perc)
{
    fprintf(stderr, "%-3d%%\r", perc);
}

void MidiScanVolume::error(const char* msg)
{
  printf("Error: %s\n", msg);
  exit(1);
}

void MidiScanVolume::warning(const char* msg)
{
  printf("Warning: %s\n", msg);
}


class MidiVolume : public MidiCopy
{
public:
  MidiVolume(char* name);

  virtual void track(int trackno, long length, int channel);

  virtual void noteon(int channel, int note, int vel);
  virtual void noteoff(int channel, int note, int vel);
  virtual void volume(int channel, int vol);

  virtual void error(const char* msg);
  virtual void warning(const char* msg);


  virtual void percent(int perc);

  int volume_;
  int scaledvolume_;
};

MidiVolume::MidiVolume(char* name) : MidiCopy(name)
{
  volume_ = 90; // default GM volume
}

void MidiVolume::track(int trackno, long length, int channel)
{
  MidiCopy::track(trackno, length, channel);
}

void MidiVolume::volume(int channel, int vol)
{
  volume_ = vol;
  if (maxlevel == 0)
    scaledvolume_ = vol;
  else
  {
    scaledvolume_ = (int)((long(vol) * expectedlevel) / maxlevel);
    if (scaledvolume_ > 127)
      scaledvolume_ = 127;
  }
  MidiCopy::volume(channel, scaledvolume_);
}

void MidiVolume::noteon(int channel, int note, int vel)
{
  vel = scalevel(volume_, vel, scaledvolume_);
  MidiCopy::noteon(channel, note, vel);
}

void MidiVolume::noteoff(int channel, int note, int vel)
{
  MidiCopy::noteoff(channel, note, vel);
}

void MidiVolume::percent(int perc)
{
    fprintf(stderr, "%-3d%%\r", perc);
}

void MidiVolume::error(const char* msg)
{
  printf("Error: %s\n", msg);
  exit(1);
}

void MidiVolume::warning(const char* msg)
{
  printf("Warning: %s\n", msg);
}

void usage()
{
  fprintf(stderr, "midivol tries to change volume of midi files to a common level\n");
  fprintf(stderr, "usage: midivol [-level #] file.mid newfile.mid\n");
  fprintf(stderr, "-version\tget program version\n");
  fprintf(stderr, "-level #\tlevel are between 1 and 16000 (default is %d)\n", expectedlevel);
  exit(1);
}

int main(int argc, char**argv)
{
  argc--; argv++;
  while (argc > 0 && **argv == '-')
  {
    if (strncmp(*argv, "-version", 2) == 0)
    {
      fprintf(stderr, "%s\n", version);
      argc--; argv++;
      if (argc == 0)
	return 0;
      continue;
    }
    if (argc > 0 && strncmp(*argv, "-level", 2) == 0)
    {
      argc--; argv++;
      if (argc > 0)
      {
	if (isdigit(**argv))
	  expectedlevel = atoi(*argv);
	else
	  fprintf(stderr, "invalid option -level parameter %s (value expected)\n", *argv);
	argc--; argv++;
      }
      else
      {
	fprintf(stderr, "missing parameter of option -level\n");
      }
      continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
    usage();
  }
  if (argc < 1)
    usage();

  input = argv[0];
  if (argc >= 2)
  {
    output = argv[1];
    if (strcmp(input, output) == 0)
    {
      fprintf(stderr, "cannot convert midi to same file\n");
      return 1;
    }
  }
  MidiScanVolume midi(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }
  midi.run();
  maxlevel = midi.maxlevel_;
  printf("%s:\nminimum level=%d\nmaximum level=%d\naverage level=%d\nmaximum volume=%d\n",
    input, midi.minlevel_, midi.maxlevel_, midi.avglevel_, midi.maxvolume_);

  if (argc == 1)
    return 0;

  MidiVolume midicopy(input);
  if (!midi.getf())
  {
    perror(input);
    return 1;
  }

  MidiWrite* write = new MidiWrite(output);
  if (!write)
  {
    fprintf(stderr, "out of memory\n");
    return 1;
  }
  if (!write->getf())
  {
    perror(output);
    return 1;
  }
  argc--; argv++;

  midicopy.setoutput(write);

  midicopy.options_ = OPTION_NOSYSEVENTS;
  if (!midicopy.run())
    fprintf(stderr, "%s: midi read error at %04lX\n", input, midi.getpos());

  delete write;
  return 0;
}
